import re
from logging import getLogger
from typing import List, Dict, Iterable

from django.apps import AppConfig
from django.urls import path
from django.utils.functional import cached_property
from django.utils.text import slugify
from gdaps.api import Interface

from medux.common.api.http import HttpResponseEmpty
from medux.common.htmx.mixins import HtmxResponseMixin

logger = getLogger(__file__)


class MeduxPluginAppConfig(AppConfig):
    """Common base class for all MedUX AppConfigs.

    All MedUX apps' AppConfigs must inherit from this class (or must at least implement this interface).
    """

    from django.db import models

    # A dict that defines groups and their permissions this Plugin sets.
    #   You can call :class:`medux.common.tools.create_groups_permissions()`
    #   when the application is initialized.
    #   Example:
    #   groups_permissions = {
    #       'Group name': {
    #           SomeModel_or_dotted_model_path: ['add', 'change', 'delete', 'view'],
    #           ...
    #       }
    #   }
    #   Note: the "Group name" is translatable, just set the English name here.
    #   You can use a model class or a dotted model path like "auth.user"
    groups_permissions: Dict[str, Dict[models.Model, List[str]]] = {}

    def initialize(self):
        """Initializes the application at setup time.

        This method is called from the "initialize" management command.
        It should set up basic data in the database etc., and needs to be idempotent.
        """

    @cached_property
    def compatibility_errors(self) -> List[str]:
        """checks for compatibility issues that can't be ignored for correct application function,
         and returns a list of errors.

        :returns a list of error strs"""

        return []

    @cached_property
    def compatibility_warnings(self) -> List[str]:
        """Checks for compatibility issues that can be accepted for continuing.

        :return: a list of warnings"""

        return []


NON_CALLABLE_ATTRIBUTES = [
    "weight",
    "separator",
    "required_permissions",
    "view_name",
    "exact_url",
]
INTERNAL_ATTRIBUTES = [
    "menu",
    "url",
    "slug",
    "title",
    "weight",
    "icon",
    "separator",
    "required_permissions",
    "view_name",
    "badge",
    "disabled",
    "exact_url",
    "check",
    "visible",
    "collapsed",
]


@Interface
class IMenuItem:
    """An extendable and versatile MenuItem Interface

    You can use that for creating menu items in a named menu:
    .. code-block: python

        class AddUserAction(IMenuItem)
            menu = "page_actions"
            title = _("Add user")
            url = reverse_lazy("user:delete")
            icon = "user-delete"

    :param menu str: The menu name where this MenuItem should be rendered. Can be any string.
        This menuitem is then found in templates under the ``menu.<this-menu-attr>`` menu.
        If you don't specify a menu, "main" is used.
    """

    __service__ = False
    menu: str = "main"
    url: str = ""
    slug: str = ""
    title: str = ""
    weight: int = 0
    icon: str = None
    separator: bool = False
    required_permissions: list = []
    view_name: str = ""
    badge = None
    disabled: bool = False
    exact_url = False
    check: bool = True
    visible: bool = True
    collapsed: bool = True

    def __init__(self, request):
        """initializes a menu item during one request"""
        self.request = request
        self._children = []
        # check permissions, and set visible as needed
        if self.required_permissions:
            if isinstance(self.required_permissions, str):
                self.required_permissions = [self.required_permissions]
            if not request.user.has_perms(self.required_permissions):
                self.visible = False
                return
        if self.view_name and not request.resolver_match.view_name == self.view_name:
            self.visible = False
            return

        # we weirdly have to iterate over the __dict__ of self's Meta object,
        # as self.__dict__ is not available in Python for GDAPS plugins...?!?
        # for attr in self.__dict__:
        #     if attr in NON_CALLABLE_ATTRIBUTES:
        for attr in ["title", "url", "badge", "icon", "check"]:
            # if attribute is callable, call it with current request as param
            if callable(getattr(self, attr)):
                # call it as static method to avoid "self" as parameter
                setattr(self, attr, getattr(self.__class__, attr)(request))

        if not self.check:
            self.visible = False
            return
        # create slug form title if not availble
        if not self.slug:
            self.slug = slugify(self.title)

    def has_children(self) -> bool:
        if self._children:
            return True
        found = False
        for item in IMenuItem:
            if (
                item.menu == self.menu and "." in item.slug
            ):  # TODO improve children handling
                parts = item.slug.split(".")
                if parts[0] == self.slug:
                    self._children.append(item(self.request))
                    found = True
        self._children.sort(key=lambda item: item.weight)
        return found

    def children(self) -> Iterable:
        if self.has_children():
            for item in self._children:
                yield item
        else:
            return []

    def selected(self) -> bool:
        """check current url against this item"""
        is_current = False
        if self.exact_url:  # FIXME: exact_url does not work
            if re.match(f"{self.url}$", self.request.path):
                is_current = True
        elif re.match(f"{self.url}", self.request.path):
            is_current = True
        return is_current

    def _underline_to_hyphen(self, attr: str):
        return attr.replace("_", "-")

    def attrs(self):
        attr_str = ""
        for attr in type(self).__dict__:
            if not attr.startswith("_") and attr not in INTERNAL_ATTRIBUTES:
                attr_str += f" {self._underline_to_hyphen(attr)}={getattr(self, attr)}"
        return attr_str

    def __getattr__(self, item):
        """For all attrs that are requested in the template and are
        not defined in the class, don't produce an error, just return an empty
        string."""
        return ""


class IHTMXComponentMixin(HtmxResponseMixin):
    """An interface Mixin that describes a component.

    Use this mixin to create interfaces for HTMX components.
    Implementations of that created interface should inherit from
    django.views.View, e.g.:
    ```
    IHTMXComponentMixin
      IMyComponent(IHTMXComponentMixin)
        MyComponentView(IMyComponent, UpdateView)
    ```

    ```python
    @Interface
    class IUserProfileSection(IHTMXComponentMixin, ...):
        params = ["pk"]
    ```

    You can also inherit from other mixins, e.g. PermissionRequiredMixin, etc.

    ```python

    class UserprofilePasswordSection(IUserProfileSection, TemplateView)
        name = "password"
        template_name = "my_app/password_view.html
    ```
    """

    params: list[str] = []
    """A list of params the view is using. These must then be passed when calling
    the view from a template, e.g. via hx-get."""

    title: str = ""
    """the title this plugin is rendered with. It's up to you how the plugin uses that title."""

    icon: str = ""
    """the icon name, if the component is listed in a list, and icons are used."""

    weight: int = 0
    """The weight this component is ranked like menu items. The more weight, the more the component sinks
    "down" in the list."""

    # HTMX is always enforced for components
    enforce_htmx = True

    def get_path(self):
        if not self.params:
            raise AttributeError(f"{self.__class__} must declare a 'params' attribute.")

        params = [f"<{p}>" for p in self.params]
        return path(
            f"{'/'.join(params)}/{self.name}/",
            self.__class__.as_view(),
            name=self.name,
        )

    def url(self):
        return f"components:{self.name}"

    @classmethod
    def get_url_patterns(cls):
        """Convenience method to add to your urls.py in an include section:

        .. code--block:: python

            path(...),
            path("profile/", include((IUserProfileSectionView.get_url_patterns(), "profile"), namespace="profile")),

        or directly add the url_patterns to the main list:

        .. code-block:: python

            url_patterns += IUserProfileSectionView.get_url_patterns()
        """
        pattern_list = []
        for plugin in cls:
            pattern = plugin.get_path()
            if not pattern:
                raise AttributeError(f"{plugin} must declare a 'url_pattern'.")
            pattern_list.append(pattern)
        return pattern_list

    # def get_context_data(self, **kwargs):
    #     """Ensures that the current view
    #     """
    #     kwargs = super().get_context_data()
    #     kwargs["object"] = self.request.user
    #     return kwargs
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({"request": self.request})
        return kwargs


class UseComponentMixin:
    """A mixin that can be added to a class that uses HTMX components.

    :param components: a list of Interfaces that are used in the template
        of this view, and can be accessed there using ``components.IFooInterface``.
    """

    components: list = []

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        component_dict = {}
        if not self.components:
            raise AttributeError(
                f"{self.__class__.__name__} has no 'components' attribute."
            )
        for component in self.components:
            component_dict[component.__name__] = list(IUserProfileSection)
        context.update({"components": component_dict})
        return context


@Interface
class IUserProfileSection(IHTMXComponentMixin):
    """Plugin hook that is rendered as HTMX view in the user profile as section."""

    template_name = "common/user_profile_section.html"
    url_pattern = None
    params = ["pk"]

    def has_permission(self):
        return self.request.user.is_authenticated

    def get_success_url(self):
        return self.request.path


class ModalFormViewMixin:
    """Mixin for FormViews that should live in a modal.

    It relies on crispy-forms intensively, and already provides a form
    helper instance attribute you can use.

    In your template, you should extend "common/modal-form.html", this
    template uses a ``header`` with a ``title`` block, a ``body``,
    and a ``footer`` block to override, for your modal dialog. In the
    footer, there is always a "Cancel" button, and as default, a "Save"
    button, which you can override using the "footer" block.

    When the modal pops up, the focus is set to the first visible input
    element.

    If the form is saved successfully, it returns an empty page
    and emits the event specified in ``success_event`` on the client,
    so that it can reload changed content.
    """

    #: A Javascript event that is triggered on the client after the form
    #: is saved successfully
    success_event: str = ""

    #: The title of the modal form
    modal_title: str = ""

    def get_modal_title(self) -> str:
        """returns: a string that is used as title of the modal."""
        return self.modal_title

    def get_success_url(self):
        return ""

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"modal_title": self.get_modal_title()})
        return context

    def form_valid(self, form):
        # call the super class, but return another Response
        super().form_valid(form)
        return HttpResponseEmpty(headers={"HX-Trigger": self.success_event})



from django.views.generic import DetailView, CreateView

from medux.core.forms import CreatePrescriptionForm
from medux.core.models import Dosage, MedicationRequest


class PrescriptionDetailView(DetailView):
    model = Dosage


class CreatePrescriptionView(CreateView):
    model = MedicationRequest
    # form_class = CreatePrescriptionForm
    fields = "__all__"



from crispy_forms.helper import FormHelper
from crispy_forms.layout import (
    Layout,
    Row,
    Column,
    ButtonHolder,
    Submit,
    LayoutObject,
    HTML,
)
from django import forms
from django.contrib.auth import password_validation
from django.contrib.auth.forms import (
    UserCreationForm as DjangoUserCreationForm,
    UsernameField,
)
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.forms import ModelForm
from django.utils.translation import gettext_lazy as _

from medux.common.bootstrap import Card
from medux.common.constants import USERS_GROUP_NAME
from medux.common.models import User, Tenant
from .mixins import ErrorLogMixin


class SignUpForm(DjangoUserCreationForm):
    username = forms.CharField(
        widget=forms.TextInput(
            attrs={"placeholder": _("Username"), "class": "form-control"}
        )
    )
    email = forms.EmailField(
        widget=forms.EmailInput(
            attrs={"placeholder": _("Email"), "class": "form-control"}
        )
    )
    password1 = forms.CharField(
        widget=forms.PasswordInput(
            attrs={
                "placeholder": _("Password"),
                "class": "form-control",
                "autocomplete": "new-password",
            }
        )
    )
    password2 = forms.CharField(
        widget=forms.PasswordInput(
            attrs={
                "placeholder": _("Password check"),
                "class": "form-control",
                "autocomplete": "new-password",
            }
        )
    )

    class Meta:
        model = User
        fields = ("username", "email", "password1", "password2")


class UserCreationForm(ErrorLogMixin, DjangoUserCreationForm):
    class Meta:
        model = User
        localized_fields = "__all__"
        fields = [
            "tenant",
            "username",
            "title",
            "first_name",
            "last_name",
            "email",
            "is_active",
            "is_superuser",
            "is_staff",
            "groups",
        ]
        field_classes = {"username": UsernameField}

    groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.all(),
        widget=forms.CheckboxSelectMultiple,
    )

    def __init__(self, **kwargs):
        self.user = kwargs.pop("user")
        super().__init__(**kwargs)
        self.fields["tenant"].initial = self.user.tenant
        self.fields["is_staff"].initial = True
        self.fields["groups"].initial = [Group.objects.get(name="Users")]
        # self.fields["date_joined"].widget = forms.HiddenInput()

        # self.fields["date_joined"].disabled = True

        self.helper = FormHelper()
        self.helper.layout = Layout(
            Card(
                Row(Column("username"), Column("tenant")),
                Row(
                    Column("title", css_class="col-2"),
                    Column("first_name"),
                    Column("last_name"),
                ),
                Row(Column("email")),
                Row(Column("password1")),
                Row(Column("password2")),
                Row(
                    Column("groups"),
                    Column("is_active"),
                    Column("is_staff"),
                    Column("is_superuser"),
                ),
                # title=_("Username: {username}").format(username=self.user.username),
                # subtitle=_("Created at: {date_joined}").format(
                #     date_joined=strftime(
                #         self.user.date_joined.date(), settings.SHORT_DATE_FORMAT
                #     )
                # ),
            ),
            ButtonHolder(
                Submit("submit", _("Save")),
                Submit("delete", _("Delete"), css_class="btn-danger"),
            ),
        )

        # self.helper.add_input(Button("cancel", _("Cancel"), css_class="btn-secondary"))

    def clean_groups(self):
        if not self.cleaned_data["groups"]:
            return [Group.objects.get(name=USERS_GROUP_NAME)]
        return self.cleaned_data["groups"]


# class AddLayout(Layout):
#     """Crispy Layout with enabled += operator"""
#
#     # FIXME: https://github.com/django-crispy-forms/django-crispy-forms/issues/1307
#     def __iadd__(self, other) -> None:
#         self.fields.append(other)
#
#     def __add__(self, other: LayoutObject) -> "AddLayout":
#         new = copy(self)
#         new += other
#         return new


class UserProfileSectionFormMixin:
    """Mixin for all User profile sections.

    It provides a ``helper`` attribute, which is a crispy FormHelper()
    instance to your usage - don't declare another form helper.
    The helper's layout is auto-generated from the models fields as provided.
    If a ``layout`` attr is given, this layout is taken.
    If a ``buttons`` attr is given, these buttons are appended to the form.
    As default, a "Save" button is defined.

    The form requires the current `request` as kwarg.
    """

    #: the crispy layout of the form. If empty, the usual form fields are taken.
    layout: Layout = None

    #: The buttons at the bottom of the form. A "Save" button as default.
    buttons: list[LayoutObject] = [Submit("submit", _("Save"))]

    def __init__(self, request, *args, **kwargs) -> None:
        self.request = request
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.layout = self.get_layout()
        super().__init__(*args, **kwargs)

    def get_layout(self) -> Layout:
        """return the declared layout, or if none, build a standard Layout
        of the original field names of the form class."""
        if not self.layout:
            self.layout = Layout(*[field for field in self.base_fields])

        return Layout(*self.layout.fields + [ButtonHolder(*self.buttons)])


class UserProfileMasterDataForm(UserProfileSectionFormMixin, ModelForm):
    class Meta:
        model = User
        localized_fields = "__all__"
        fields = [
            # "username",
            "tenant",
            "title",
            "first_name",
            "last_name",
            "email",
            "avatar",
        ]

    layout = Layout(
        Row(
            Column("title", css_class="col-2"),
            Column("first_name"),
            Column("last_name"),
        ),
        Row(Column("email"), Column("tenant")),
        Row(
            Column("avatar"),
            Column(
                HTML(
                    """{% load static %}
                    <span class='avatar avatar-xl mb-3'
                    style='background-image: url({% if object.avatar %}{{object.avatar.url}}{% else %}
                    {% static "images/default_avatar.png" %}{% endif %})'>
                    """
                ),
            ),
        ),
        # subtitle=_("Created at: {date_joined}").format(
        #     date_joined=strftime(
        #         self.user.date_joined.date(), settings.SHORT_DATE_FORMAT
        #     )
        # ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.request.user.is_superuser:
            self.fields["tenant"].queryset = Tenant.objects.filter(
                id=self.instance.tenant.id
            )
            self.fields["tenant"].disabled = True


class UserProfilePasswordForm(UserProfileSectionFormMixin, forms.ModelForm):
    """
    A form that lets a user change set their password without entering the old
    password.
    """

    class Meta:
        model = User
        fields = ["id"]

    # FIXME - better inherit SetPasswordForm directly - but that must be fixed first.
    # This view is copied from django.contrib.auth.forms.SetPasswordForm,
    # as this form needs a "user" argument instead of "instance", this
    # makes it incompatible with other UserProfileSectionForms
    error_messages = {
        "password_mismatch": _("The two password fields didn’t match."),
    }
    new_password1 = forms.CharField(
        label=_("New password"),
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        strip=False,
        help_text=password_validation.password_validators_help_text_html(),
    )
    new_password2 = forms.CharField(
        label=_("New password confirmation"),
        strip=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
    )

    def __init__(self, user, *args, **kwargs):
        # this method gets a "user" attr from the PasswordChangeView, and
        # a "instance" kwarg from UserProfileSectionFormMixin,
        # but does not inherit ModelForm, so it can't pass them to its parents
        self.user = user
        kwargs.pop("instance")
        super().__init__(*args, **kwargs)

    def clean_new_password2(self):
        password1 = self.cleaned_data.get("new_password1")
        password2 = self.cleaned_data.get("new_password2")
        if password1 and password2:
            if password1 != password2:
                raise ValidationError(
                    self.error_messages["password_mismatch"],
                    code="password_mismatch",
                )
        password_validation.validate_password(password2, self.user)
        return password2

    def save(self, commit=True):
        password = self.cleaned_data["new_password1"]
        self.user.set_password(password)
        if commit:
            self.user.save()
        return self.user


class UserProfileGroupsForm(UserProfileSectionFormMixin, ModelForm):
    class Meta:
        model = User
        fields = ["groups", "is_active"]

    layout = Layout(Row(Column("groups"), Column("is_active")))

    groups = forms.ModelMultipleChoiceField(
        widget=forms.CheckboxSelectMultiple(), queryset=Group.objects.all()
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.fields["is_active"].disabled = not self.request.user.is_superuser
        if not self.request.user.is_superuser:
            self.fields["is_active"].widget = forms.HiddenInput()



# empty test so that pytest doesn't fail
# remove this test when you created your own
import pytest
from django.contrib.auth import get_user_model

from medux.preferences.definitions import Scope, KeyType
from medux.preferences.loaders import PreferencesLoader
from medux.preferences.registry import PreferencesRegistry


@pytest.mark.django_db
def test_load_key():
    PreferencesRegistry.register(
        "test_namespace", "test_key", [Scope.USER], key_type=KeyType.INTEGER
    )
    User = get_user_model()
    admin = User.objects.create(username="admin")
    admin.save()
    loader = PreferencesLoader(scope=Scope.USER, foreign_object=admin)
    loader.update_preferences({"test_namespace": {"test_key": 42}})


@pytest.mark.django_db
def test_load_key_without_registering():
    loader = PreferencesLoader(scope=Scope.VENDOR)
    with pytest.raises(KeyError):
        loader.update_preferences({"notexisting_namespace": {"new_key": 42}})



from django_unicorn.components import UnicornView


class MedicationrequestEntryView(UnicornView):
    med_text = ""
    is_editing: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.med_text = kwargs.get("med_text")

    def search(self):
        self.med_text = self.med_text.upper()
        self.is_editing = False

    def clear(self):
        self.med_text = ""



# This file contains base models for master data used in MedUX. They can be provided using packages.

import uuid
from django.db import models
from django.utils.translation import gettext_lazy as _

from medux.common.models import BaseModel
from medux.core.fields import MarkdownField


class PackageDataModel(models.Model):
    """An abstract base model for everything that can be packaged.

    .. note:

        This is not inheriting :class:`BaseModel`, so no soft-delete is available here.
    """

    USER = 1  # User generated data
    LOCAL_ADMIN = 10
    VENDOR = 99
    MAINTAINER = (
        (USER, _("User")),
        (LOCAL_ADMIN, _("Admin")),
        (VENDOR, _("Vendor")),
    )
    # the uuid field is for uniquely identifying this item across practices, to simplify distribution
    uuid = models.UUIDField(default=uuid.uuid4)

    maintainer = models.SmallIntegerField(choices=MAINTAINER, default=USER)

    # TODO: include a "dirty" field for indicating that the user, admin etc. has changed something
    # So it can be excluded from updating, depending on what the user wants.

    class Meta:
        abstract = True


class DataPack(BaseModel):
    """A data pack that can be downloaded from an update server.

    A data pack contains data in the form of :class:`PackageDataModel`s.
    This is basically what a ValueSet in FHIR means.

    A DataPack is intended to be downloaded from an update server as zip file.
    The metadata can be saved in this DataPack model, to keep a trace of it."""

    uuid = models.UUIDField(default=uuid.uuid4)
    name = models.CharField(
        max_length=100, help_text=_("Name for this data pack (computer friendly)")
    )
    title = models.CharField(
        max_length=255, help_text=_("Name for this data pack (human friendly)")
    )
    description = MarkdownField(
        max_length=255, help_text=_("Natural description, in MarkDown")
    )
    license = models.CharField(max_length=20)
    version = models.CharField(max_length=25)
    experimental = models.BooleanField(default=False)
    publisher = models.CharField(max_length=255, blank=True)
    language = models.ForeignKey("Language", on_delete=models.PROTECT)
    # TODO medux compatibility version

    #: the model in which the field should be saved
    model = models.CharField(max_length=255)
    downloaded = models.BooleanField(default=False)
    installed = models.BooleanField(default=False)
    data_file = models.FileField(upload_to="datapacks")



from django.apps import apps
from django.conf import settings
from django.contrib import admin


# taken from https://adriennedomingus.medium.com/soft-deletion-in-django-e4882581c340
from django.contrib.auth.models import Group

from .forms import DosageForm
from .models import (
    Name,
    Address,
    Country,
    ContactPoint,
    AdministrativeGender,
    HealthServiceProvider,
    Physician,
    Hospital,
    Person,
    Patient,
    Encounter,
    Problem,
    Narrative,
    NarrativeType,
    Department,
    Specialty,
    Language,
    Dosage,
)


class MeduxAdmin(admin.ModelAdmin):
    """A ModelAdmin for MeduxModels with Softdeletion awareness"""

    def get_queryset(self, request):
        # use the all_objects manager
        qs = self.model.all_objects

        # The below is copied from the base implementation in BaseModelAdmin to
        # prevent other changes in behavior
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs

    def delete_model(self, request, obj):
        obj.hard_delete()


class AdministrativeGenderAdmin(admin.ModelAdmin):
    ordering = ("sort_weight",)


admin.site.register(AdministrativeGender, AdministrativeGenderAdmin)


class DosageAdmin(admin.ModelAdmin):
    form = DosageForm
    fields = [
        "method",
        "dose_type",
        "dose_rate_numerator_value",
        "dose_rate_numerator_unit",
    ]


admin.site.register(Dosage, DosageAdmin)

for model in [
    Country,
    ContactPoint,
    Address,
    # Period,
    HealthServiceProvider,
    Physician,
    Specialty,
    Hospital,
    Person,
    Patient,
    Encounter,
    Problem,
    Narrative,
    NarrativeType,
    Name,
    Department,
    Language,
]:
    admin.site.register(model)


if settings.DEBUG:
    # all other models
    models = apps.get_models()

    for model in models:
        try:
            if not model._meta.abstract:
                admin.site.register(model)
        except admin.sites.AlreadyRegistered:
            pass



from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.http import HttpRequest

from medux.common.models import Tenant
from medux.preferences import Scope, CachedSettings
from medux.preferences.models import ScopedPreference
from medux.preferences.registry import PreferencesRegistry
from medux_online.core.tests import MeduxOnlineTestCase

User = get_user_model()


class CachedSettingsTest(MeduxOnlineTestCase):
    def setUp(self) -> None:
        # create a dummy request
        self.request = HttpRequest()

    def test_cachedpreferences_namespace_type(self):
        namespace = CachedSettings(self.request).mynamespace
        self.assertEqual(type(namespace), CachedSettings.Namespace)

    def test_cachedpreferences_key_None(self):
        key = CachedSettings(self.request).test.non_existing_key
        self.assertIsNone(key)

    def test_cachedpreferences_existing_key_vendor(self):
        PreferencesRegistry.register("space", "my_key")
        ScopedPreference.set("space", "my_key", 42, Scope.VENDOR)
        self.assertEqual(CachedSettings(self.request).space.my_key, 42)

    def test_cachedpreferences_existing_key_tenant(self):
        PreferencesRegistry.register("space", "my_key", [Scope.TENANT])
        tenant = Tenant.objects.create(
            first_name="Demo", last_name="User", sex="f", address="Foo street"
        )
        tenant.save()

        ScopedPreference.set("space", "my_key", 42, Scope.TENANT, tenant=tenant)
        self.request.site = Site.objects.get(pk=1)
        space = CachedSettings(self.request).space
        key = space.my_key
        assert key == 42

    def test_cachedpreferences_existing_key_with_correct_user(self):
        user1 = User.objects.create_user(username="user1", password="user1")
        PreferencesRegistry.register("space", "my_key", [Scope.USER])
        ScopedPreference.set("space", "my_key", 42, Scope.USER, user=user1)
        self.request.user = user1
        self.assertEqual(CachedSettings(self.request).space.my_key, 42)

    def test_cachedpreferences_existing_key_with_wrong_user(self):
        user1 = User.objects.create_user(username="user1", password="user1")
        user2 = User.objects.create_user(username="user2", password="user2")

        ScopedPreference.set("space", "my_key", 42, Scope.USER, user=user1)
        self.request.user = user2
        self.assertIsNotNone(CachedSettings(self.request).space)

        key = CachedSettings(self.request).space.my_key
        self.assertIsNone(key)

    def test_cachedpreferences_existing_key_user_fallback_vendor(self):
        user1 = get_user_model().objects.create(username="user1", password="user1")
        user2 = get_user_model().objects.create(username="user2", password="user2")
        PreferencesRegistry.register("space", "my_key", [Scope.USER])

        ScopedPreference.set("space", "my_key", 41, Scope.VENDOR)
        ScopedPreference.set("space", "my_key", 42, Scope.USER, user=user1)
        self.request.user = user2
        self.assertEqual(CachedSettings(self.request).space.my_key, 41)



import pytest

from medux.core.models import Name


@pytest.mark.django_db
def test_name_create():
    name = Name(first_name="Tom", last_name="Green")
    name.save()

    assert name.id
    assert name.first_name == "Tom"
    assert name.last_name == "Green"
    assert str(name) == "GREEN, Tom"
    assert name.weight == 1



"""
Django settings for medux project.

Generated by 'django-admin startproject' using Django 3.1.5.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

import environ
from pathlib import Path
from gdaps.pluginmanager import PluginManager

# Initialise environment variables
env = environ.Env(
    # set casting, default value
    # DEBUG=(bool, False)
)

# First look in CWD for .env file, use it if if exists
# This is useful for plugin development in separate directories/virtualenvs
if Path(".env").exists():
    env.read_env(env_file=str(Path.cwd() / ".env"))
else:
    # if not, use the default location from django-environ
    env.read_env()

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

PROJECT_NAME = "medux"

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env(
    "SECRET_KEY", default="don't use this pseudo-secret-key on production!"
)

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env("DEBUG", default=False)

# Beware: Django Debug Toolbar seems to have problems with Sockpuppet.
DEBUG_TOOLBAR = False


ALLOWED_HOSTS = env("ALLOWED_HOSTS", default="127.0.0.1,localhost").split(",")

# Application definition
INSTALLED_APPS = [
    "whitenoise.runserver_nostatic",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_extensions",
    "django.forms",
    "crispy_forms",
    "crispy_bootstrap5",
    "gdaps",
    "django_htmx",
    # "django_filters",
    "medux.common",
    "medux.preferences",
    "component_tags",
    "composite_field",
]

SITE_ID = 1

if DEBUG_TOOLBAR:
    INSTALLED_APPS += ["debug_toolbar"]

# Note: Plugins that are included per default within this repository
# should be added to the [options.entry_points] section in MdeUX' setup.cfg
INSTALLED_APPS += PluginManager.find_plugins("medux.plugins")

# https://docs.djangoproject.com/en/dev/ref/settings/#auth-user-model
AUTH_USER_MODEL = "common.User"

MIDDLEWARE = []
if DEBUG_TOOLBAR:
    MIDDLEWARE += ["debug_toolbar.middleware.DebugToolbarMiddleware"]

MIDDLEWARE += [
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.contrib.sites.middleware.CurrentSiteMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
    "medux.core.middleware.DeviceMiddleware",
]

ROOT_URLCONF = "medux.urls"

# https://docs.djangoproject.com/en/dev/ref/settings/#login-url
LOGIN_URL = "/accounts/login/"
LOGIN_REDIRECT_URL = "home"
LOGOUT_REDIRECT_URL = "home"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
            "builtins": ["component_tags.template.builtins"],
        },
    },
]
TEMPLATE_CONTEXT_PROCESSORS = ["django.core.context_processors.request"]
# https://docs.djangoproject.com/en/3.2/ref/settings/#std:setting-FORM_RENDERER
FORM_RENDERER = "django.forms.renderers.TemplatesSetting"


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

DATABASES = {
    "default": {
        "ENGINE": env("DATABASE_ENGINE", default="django.db.backends.sqlite3"),
        "NAME": env("DATABASE_NAME", default="medux.sqlite3"),
        "USER": env("DATABASE_USER", default=""),
        "PASSWORD": env("DATABASE_PASS", default=""),
    }
}

# https://docs.djangoproject.com/en/3.2/releases/3.2/#customizing-type-of-auto-created-primary-keys
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = "de-AT"

TIME_ZONE = "Europe/Vienna"

USE_I18N = True

USE_L10N = True

USE_TZ = True

LOCALE_PATHS = [BASE_DIR / "locale"]

# Static files (CSS, JavaScript, Images)
STATIC_URL = "/static/"
STATIC_ROOT = BASE_DIR / "static"
STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"

MEDIA_ROOT = BASE_DIR / "media"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler"}},
    "loggers": {
        "gdaps": {"handlers": ["console"], "level": "INFO", "propagate": True},
        "django": {"handlers": ["console"], "level": "INFO", "propagate": True},
        # "django.template": {
        #     "handlers": ["console"],
        #     "level": "DEBUG",
        #     "propagate": True,
        # },
        "unicorn": {"handlers": ["console"], "level": "INFO"},
        "medux": {"handlers": ["console"], "level": "DEBUG"},
    },
}


# GDAPS configuration
GDAPS = {}

CHANNEL_LAYERS = {
    "default": {
        # "BACKEND": "channels.layers.InMemoryChannelLayer",
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [("127.0.0.1", 6379)],
        },
    },
}

ASGI_APPLICATION = "medux.asgi.application"

# https://django-debug-toolbar.readthedocs.io/en/latest/installation.html
INTERNAL_IPS = ["127.0.0.1", "localhost"]

if DEBUG_TOOLBAR:
    DEBUG_TOOLBAR_PANELS = [
        "debug_toolbar.panels.history.HistoryPanel",
        "debug_toolbar.panels.versions.VersionsPanel",
        "debug_toolbar.panels.timer.TimerPanel",
        "debug_toolbar.panels.settings.SettingsPanel",
        "debug_toolbar.panels.headers.HeadersPanel",
        # 'debug_toolbar.panels.request.RequestPanel',  # remove due to problems with ASGI
        "debug_toolbar.panels.sql.SQLPanel",
        "debug_toolbar.panels.staticfiles.StaticFilesPanel",
        "debug_toolbar.panels.templates.TemplatesPanel",
        "debug_toolbar.panels.cache.CachePanel",
        "debug_toolbar.panels.signals.SignalsPanel",
        "debug_toolbar.panels.logging.LoggingPanel",
        "debug_toolbar.panels.redirects.RedirectsPanel",
        "debug_toolbar.panels.profiling.ProfilingPanel",
    ]


# REST_FRAMEWORK = {
#     "DEFAULT_PERMISSION_CLASSES": [
#       "rest_framework.permissions.DjangoModelPermissions"
#     ]
# }

MEDUX = {"BIRTHDAY_FORMAT": "Date"}

# django-crispy.forms
CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap5"
CRISPY_TEMPLATE_PACK = "bootstrap5"

if DEBUG:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    }



from django_unicorn.components import UnicornView

from medux.core.models import Patient


class PatientListView(UnicornView):
    patient_list = Patient.objects.none()

    def load_table(self, params: dict):
        self.patient_list = Patient.objects.filter(**params)



from unittest import TestCase

from medux.preferences import Scope, KeyType
from medux.preferences.registry import PreferencesRegistry


class TestRegistry(TestCase):
    def test_register_key(self):
        PreferencesRegistry.register("namespace", "key1", [Scope.USER, Scope.VENDOR])
        self.assertTrue(PreferencesRegistry.exists("namespace", "key1", Scope.USER))

    def test_register_wrong_key(self):
        PreferencesRegistry.register("namespace", "key2", [Scope.USER, Scope.VENDOR])
        self.assertFalse(PreferencesRegistry.exists("namespace", "key33", Scope.USER))

    def test_register_wrong_namespace(self):
        PreferencesRegistry.register("namespace", "key3", [Scope.USER, Scope.VENDOR])
        self.assertFalse(PreferencesRegistry.exists("foospace", "key45", Scope.USER))

    def test_register_wrong_scope(self):
        PreferencesRegistry.register("namespace", "key4", [Scope.USER, Scope.VENDOR])
        self.assertFalse(PreferencesRegistry.exists("namespace", "key4", Scope.DEVICE))

    def test_register_implicitly_vendor_scope(self):
        """check if registered settings add VENDOR scope automatically"""
        PreferencesRegistry.register("namespace", "key5", [Scope.USER])
        self.assertTrue(PreferencesRegistry.exists("namespace", "key5", Scope.VENDOR))

    def test_retrieve_scopes(self):
        PreferencesRegistry.register("namespace", "key6", [Scope.USER, Scope.DEVICE])
        self.assertEqual(
            PreferencesRegistry.scopes("namespace", "key6"),
            {
                Scope.USER,
                Scope.DEVICE,
                Scope.VENDOR,
            },
        )

    def test_retrieve_empty_scopes(self):
        PreferencesRegistry.register("namespace", "key_empty", [])
        self.assertEqual(
            PreferencesRegistry.scopes("namespace", "key_empty"), {Scope.VENDOR}
        )

    def test_retrieve_vendor_scopes(self):
        PreferencesRegistry.register("namespace", "key_vendor", [Scope.VENDOR])
        self.assertEqual(
            PreferencesRegistry.scopes("namespace", "key_vendor"), {Scope.VENDOR}
        )

    def test_all(self):
        PreferencesRegistry.register(
            "namespace", "key_vendor", [Scope.VENDOR], key_type=KeyType.INTEGER
        )
        PreferencesRegistry.register("namespace2", "key_user", [Scope.USER])
        result = list(PreferencesRegistry.all())
        # self.assertEqual(len(result), 3)
        self.assertIn(
            ("namespace", "key_vendor", Scope.VENDOR, KeyType.INTEGER), result
        )
        self.assertIn(("namespace2", "key_user", Scope.VENDOR, KeyType.STRING), result)
        self.assertIn(("namespace2", "key_user", Scope.USER, KeyType.STRING), result)



#  MedUX - Open Source Electronical Medical Record
#  Copyright (c) 2022  Christian González
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from django.utils.translation import gettext_lazy as _

from medux.common.api import MeduxPluginAppConfig
from . import __version__

logger = logging.getLogger(__file__)


class PreferencesConfig(MeduxPluginAppConfig):
    name = "medux.preferences"
    verbose_name = "Settings"
    default = True  # FIXME: Remove when django bug is fixed
    groups_permissions = {
        "Users": {"preferences.ScopedPreference": ["view", "change_own_user"]},
        "Tenant admins": {"preferences.ScopedPreference": ["change_own_tenant"]},
    }

    class PluginMeta:
        """This configuration is the introspection data for plugins."""

        # the plugin machine "name" is taken from the AppConfig, so no name here
        verbose_name = _("Preferences")
        author = "Christian González"
        author_email = "office@nerdocs.at"
        vendor = "Nerdocs"
        description = _(
            "MedUX preferences tools, which are used for MedUX Online and MedUX"
        )
        category = _("Base")
        visible = True
        version = __version__

    # compatibility = "medux.core>=2.3.0"



import logging

from django_unicorn.components import UnicornView

from medux.utils import parse_date_string

logger = logging.getLogger(__name__)


class PatientSearchInputView(UnicornView):
    search_text = ""

    def search(self):
        params = {}
        self.search_text = self.search_text.strip().lower()  # type: str
        if self.search_text.isnumeric():
            # FIXME: birth date search does not work yet
            params["birth_date__date"] = parse_date_string(self.search_text)
        elif self.search_text.count(",") == 1:
            (
                params["names__last_name__startswith"],
                params["names__first_name__startswith"],
            ) = self.search_text.split(",")
        else:
            params["names__last_name__startswith"] = self.search_text

        self.parent.load_table(params)
        logger.debug(params)



import logging
from pathlib import Path

from django.apps import apps
from django.core.management import BaseCommand

from medux.preferences.loaders import TomlPreferencesLoader

logger = logging.getLogger(__file__)


class Command(BaseCommand):
    """Loads vendor preferences.toml in each app into the database"""

    help = "DEV COMMAND: Load (VENDOR) preferences from a file into the database."

    def handle(self, *args, **options):
        for app in apps.get_app_configs():
            filename = Path(app.path, "preferences.toml")
            if filename.exists():
                from medux.preferences.definitions import Scope

                loader = TomlPreferencesLoader(filename, scope=Scope.VENDOR)
                loader.load()

                logger.info(f"Loaded preferences from {filename} into database.")



from django.core.validators import RegexValidator

# TODO: maybe move into a FHIR module?
fhir_server_allowed_references = (
    "Account|ActivityDefinition|AdverseEvent|AllergyIntolerance|Appointment|"
    "AppointmentResponse|AuditEvent|Basic|Binary|BiologicallyDerivedProduct|"
    "BodyStructure|Bundle|CapabilityStatement|CarePlan|CareTeam|ChargeItem|Claim|"
    "ClaimResponse|ClinicalImpression|CodeSystem|Communication|CommunicationRequest|"
    "CompartmentDefinition|Composition|ConceptMap|Condition|Consent|Contract|"
    "Coverage|DetectedIssue|Device|DeviceComponent|DeviceMetric|DeviceRequest|"
    "DeviceUseStatement|DiagnosticReport|DocumentManifest|DocumentReference|"
    "EligibilityRequest|EligibilityResponse|Encounter|Endpoint|EnrollmentRequest|"
    "EnrollmentResponse|EntryDefinition|EpisodeOfCare|EventDefinition|"
    "ExampleScenario|ExpansionProfile|ExplanationOfBenefit|FamilyMemberHistory|Flag|"
    "Goal|GraphDefinition|Group|GuidanceResponse|HealthcareService|ImagingStudy|"
    "Immunization|ImmunizationEvaluation|ImmunizationRecommendation|"
    "ImplementationGuide|Invoice|ItemInstance|Library|Linkage|List|Location|"
    "Measure|MeasureReport|Media|Medication|MedicationAdministration|"
    "MedicationDispense|MedicationRequest|MedicationStatement|MedicinalProduct|"
    "MedicinalProductAuthorization|MedicinalProductClinicals|"
    "MedicinalProductDeviceSpec|MedicinalProductIngredient|"
    "MedicinalProductPackaged|MedicinalProductPharmaceutical|MessageDefinition|"
    "MessageHeader|NamingSystem|NutritionOrder|Observation|"
    "ObservationDefinition|OccupationalData|OperationDefinition|"
    "OperationOutcome|Organization|OrganizationRole|Patient|PaymentNotice|"
    "PaymentReconciliation|Person|PlanDefinition|Practitioner|"
    "PractitionerRole|Procedure|ProcessRequest|ProcessResponse|"
    "ProductPlan|Provenance|Questionnaire|QuestionnaireResponse|"
    "RelatedPerson|RequestGroup|ResearchStudy|ResearchSubject|"
    "RiskAssessment|Schedule|SearchParameter|Sequence|"
    "ServiceRequest|Slot|Specimen|SpecimenDefinition|StructureDefinition|"
    "StructureMap|Subscription|Substance|SubstancePolymer|"
    "SubstanceReferenceInformation|SubstanceSpecification|SupplyDelivery|"
    "SupplyRequest|Task|TerminologyCapabilities|TestReport|TestScript|"
    "UserSession|ValueSet|VerificationResult|VisionPrescription" + "|Any"
)


CodeValidator = RegexValidator(
    regex=r"[^\s]+([\s]?[^\s]+)*", message="A Code must not contain whitespaces."
)

IdValidator = RegexValidator(
    regex=r"[a-z][A-Z][0-9][-.]?",
    message="Ids can be up to 64 characters long, "
    "and contain any combination of upper "
    "and lowercase ASCII letters, numerals,"
    ' "-" and "."',
)

OidValidator = RegexValidator(r"[0-2](\.[1-9]\d*)+", message="Given string is no OID")



from django import template

from medux.preferences.definitions import Scope
from medux.preferences.models import ScopedPreference
from medux.preferences.registry import PreferencesRegistry
from medux.preferences.tools import is_effective

register = template.Library()


@register.simple_tag(takes_context=True)
def effective(context, setting):
    request = context["request"]
    return is_effective(setting, request)


@register.simple_tag
def is_setting_editable(item: ScopedPreference, request) -> bool:
    # FIXME: bool is a senseless return value in a template tag

    if not PreferencesRegistry.exists(item.namespace, item.key, item.scope):
        return False
    if item.scope == Scope.TENANT:
        return False

    if (
        item.scope == Scope.USER
        and item.user == request.user
        and request.user.has_perm("preferences.change_own_user_preferences")
    ):
        return True

    if (
        item.scope == Scope.GROUP
        and item.group in request.user.groups
        and request.user.has_perm("preferences.change_group_preferences")
    ):
        return True

    if (
        item.scope == Scope.TENANT
        and item.tenant in request.user.tenant
        and request.user.has_perm("preferences.change_own_tenant_preferences")
    ):
        return True
    return False



import logging

logger = logging.getLogger(__file__)


class ErrorLogMixin:
    """A mixin that can be added to a Form, so that it logs all form
    errors."""

    def add_error(self, field, error):
        if field:
            logger.info("Form error on field %s: %s", field, error)
        else:
            logger.info("Form error: %s", error)
        super().add_error(field, error)



from django.db import models
from django.db.models import PositiveIntegerField
from django.utils.translation import gettext_lazy as _

from medux.common.models import BaseModel
from .datapacks import PackageDataModel
from ..fields import PeriodField


# noinspection PyUnresolvedReferences
# noinspection PyAttributeOutsideInit
class UpperCodeMixin:
    def save(self, *args, **kwargs):
        self.code = self.code.upper()
        super().save(*args, **kwargs)


class Language(UpperCodeMixin, PackageDataModel):
    """A Language name like "German", "English", used all over the place in MedUX."""

    # see https://en.wikipedia.org/wiki/ISO_639-1
    code = models.CharField(
        max_length=2, primary_key=True, help_text=_("ISO 639-1 language code")
    )
    name = models.CharField(
        max_length=50, help_text=_("ISO 639-1 English language name")
    )

    def __str__(self):
        return f"{self.name} ({self.code})"


class ZipCode(PackageDataModel):
    # TODO add validator
    code = models.CharField(max_length=10)

    def __str__(self):
        return self.code


class City(PackageDataModel):
    name = models.CharField(max_length=100)
    # a city can have more than one zipcode
    zipcodes = models.ManyToManyField(ZipCode)

    class Meta:
        verbose_name_plural = _("Cities")

    def __str__(self):
        return self.name


class Country(UpperCodeMixin, PackageDataModel):
    """Countries coded per ISO 3166-1

    see https://en.wikipedia.org/wiki/ISO_3166-1"""

    code = models.CharField(max_length=2, primary_key=True, help_text="ISO 3166-1 code")
    name = models.CharField(max_length=255, help_text=_("Official English short name"))
    # the flag of the country, as Unicode char
    flag = models.CharField(max_length=5, blank=True)

    default_language = models.ForeignKey(
        Language, blank=True, null=True, on_delete=models.PROTECT
    )

    class Meta:
        verbose_name_plural = _("Countries")

    def __str__(self):
        return self.name


class AddressType(models.Model):
    """Type of an address, like home, work, parents, holidays etc."""

    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class Address(BaseModel):
    street = models.CharField(max_length=255, blank=True)

    # additional street-level information which formatters would usually
    # put on lines directly below the street line of an address, such as
    # postal box directions in CA, or c/o hints
    aux_street = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("hints like postal box directions in CA, or c/o, etc."),
    )

    number = models.CharField(
        max_length=50,
        blank=True,
        help_text=_("Street number, evtl. with additions like stair number, etc."),
    )

    # directions *below* the unit (eg.number) level, such as apartment number,
    # room number, level, entrance or even verbal directions
    subunit = models.CharField(max_length=255, blank=True)

    postcode = models.CharField(max_length=10, blank=True)

    city = models.CharField(max_length=255, blank=True)

    state = models.ForeignKey(Country, blank=True, null=True, on_delete=models.PROTECT)

    addendum = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Any additional information that did not fit anywhere else"),
    )

    location = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("The exact location of this address in latitude-longitude"),
    )

    # TODO: make sure that period gets a default value - "now -> ongoing"
    # default=ongoing_period
    period = PeriodField()

    weight = PositiveIntegerField(default=1)

    def __str__(self):
        return f"{self.street} {self.number}, {self.postcode} {self.city}"

    class Meta:
        verbose_name_plural = _("Addresses")
        unique_together = ("street", "number", "postcode", "city", "state")
        ordering = ["weight"]


class AddressMapper(models.Model):
    person_id = models.ForeignKey("Person", on_delete=models.PROTECT)
    address_id = models.ForeignKey(Address, on_delete=models.PROTECT)
    address_type = models.ForeignKey(AddressType, on_delete=models.PROTECT)



from django.urls import path, re_path, include
from django.contrib.auth import views

from django.contrib.auth.views import LoginView

from .views import HomeView, PatientFileView
from .views.medication_detail import CreatePrescriptionView
from .views.patient_file import NewPatientView, PatientListView, SettingsView

app_name = "core"

root_urlpatterns = [
    path("", HomeView.as_view(), name="home"),
    path("preferences/", SettingsView.as_view(), name="preferences"),
    path("patient/", PatientListView.as_view(), name="patient_list"),
    path("patient/file/<pk>", PatientFileView.as_view(), name="patient_file"),
    path("patient/add/", NewPatientView.as_view(), name="patient_new"),
    path("accounts/", include("django.contrib.auth.urls")),
    path(
        "medication/prescriptions/add",
        CreatePrescriptionView.as_view(),
        name="create_prescription",
    ),
]

# Django login accounts views:
# accounts/login/ [name='login']
# accounts/logout/ [name='logout']
# accounts/password_change/ [name='password_change']
# accounts/password_change/done/ [name='password_change_done']
# accounts/password_reset/ [name='password_reset']
# accounts/password_reset/done/ [name='password_reset_done']
# accounts/reset/<uidb64>/<token>/ [name='password_reset_confirm']
# accounts/reset/done/ [name='password_reset_complete']



import logging

from django.core.management import BaseCommand, call_command
from gdaps.pluginmanager import PluginManager

from medux.common import tools
from medux.common.models import User

logger = logging.getLogger(__file__)


class Command(BaseCommand):
    """Loads vendor preferences.toml in each app into the database"""

    help = "DEV COMMAND: Initializes the application for the first time. Sets up admin account, etc."

    def handle(self, *args, **options):

        admin = User.objects.filter(username="admin").first()
        assert admin
        # if admin is None:
        #     msg = "Creating Admin account: user: 'admin', password: 'admin'... "
        #     admin = User.objects.create_user(username="admin", password="admin")
        #     admin.is_staff = True
        #     admin.is_superuser = True
        #     admin.save()
        #     msg += self.style.SUCCESS("OK")
        #     self.stdout.write(msg)

        # load all apps' fixtures named "initial.[yaml|json|...]"
        call_command("loaddata", "initial")

        # Initialize plugins...
        for app in PluginManager.plugins():
            logger.info(f"Initializing {app.label}...")
            # noinspection PyUnresolvedReferences
            if hasattr(app, "initialize") and callable(app.initialize):
                app.initialize()
            logger.info(f"Creating {app.label} groups/permissions...")
            if hasattr(app, "groups_permissions"):
                # noinspection PyUnresolvedReferences
                tools.create_groups_permissions(app.groups_permissions)



from unittest import TestCase

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from medux.common.tools import create_groups_permissions

# noinspection PyPep8Naming
User = get_user_model()


@pytest.mark.django_db
class GroupPermissionsTest(TestCase):
    def setUp(self) -> None:
        self.group = Group(name="Test group 234")
        self.group.save()

        self.user = User.objects.create_user(username="demo_user", password="foo")
        self.user.groups.add(self.group)

    def test_check_user_in_group(self):
        groups = self.user.groups
        filtered = groups.filter(name=self.group)
        self.assertTrue(filtered.exists())
        # assert demo_user.groups.filter(name=group).exists()

    def test_add_permissions_to_group_with_user_as_model(self):
        groups_permissions = {"Test group 234": {User: ["add", "view"]}}
        create_groups_permissions(groups_permissions)
        self.assertEqual(len(self.group.permissions.filter(codename="view_user")), 1)
        self.assertTrue(self.group.permissions.filter(codename="view_user").exists())
        self.assertTrue(self.group.permissions.filter(codename="add_user").exists())

    def test_add_permissions_to_group_with_user_as_dotted_string(self):
        groups_permissions = {"Test group 234": {"common.User": ["add", "view"]}}
        create_groups_permissions(groups_permissions)
        self.assertTrue(self.group.permissions.filter(codename="view_user").exists())
        self.assertTrue(self.group.permissions.filter(codename="add_user").exists())



# Generated by Django 4.1 on 2022-10-17 08:37
import uuid

from django.contrib.auth import get_user_model
from django.db import migrations
from django.db.migrations import RunPython


def create_tenant(apps, schema_editor):
    Tenant = apps.get_model("common", "Tenant")
    # create a new admin tenant
    Tenant.objects.create(
        id=1,
        uuid=uuid.uuid4(),  # random Admin UID for this site.
        first_name="Admin",
        last_name="Admin",
        address="Nowhere 42, 12345 Wherever",
        sex="f",
    )


def modify_admin(apps, schema_editor):
    User = get_user_model()
    # create admin and set its tenant to newly created tenant
    User.objects.create_superuser(
        pk=1,
        username="admin",
        password="admin",
        first_name="Admin",
        last_name="Admin",
        tenant_id=1,
    )


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0001_initial"),
    ]

    operations = [
        RunPython(create_tenant),
        RunPython(modify_admin),
    ]



from crispy_forms import bootstrap
from django import forms

# ------------ Django Widgets ------------


class ButtonGroupRadioSelect(forms.RadioSelect):
    """A Bootstrap button-group input widget instead of a radio."""

    template_name = "common/widgets/buttongroup_radioselect.html"


class SwitchInput(forms.CheckboxInput):
    """A Bootstrap switch input widget instead of a checkbox."""

    template_name = "common/widgets/switch.html"

    def __init__(self, wrapper_class: str = None, **kwargs):
        super().__init__(**kwargs)
        self.wrapper_class = wrapper_class or ""

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context["widget"]["attrs"]["class"] = "form-check-input"
        context.update({"wrapper": {"class": self.wrapper_class}})
        return context


# ------------ Crispy Layout Fields ------------


class BooleanTextFieldButtonsGroup(bootstrap.Field):
    template = "%s/layout/textfield_buttonsgroup.html"



from django.test import TestCase

from medux.common.tools import create_groups_permissions

groups_permissions = {
    "Site tester": {
        "core.TermsAndConditionsPage": ["view", "add", "change", "delete"],
        "core.PrivacyPage": ["view", "add", "change"],
    },
    # "Site admins": {
    #     "core.TermsAndConditionsPage": ["view", "add", "change", "delete"],
    #     "core.PrivacyPage": ["view", "add", "change", "delete"],
    # },
}


class TestPermissionsCreation(TestCase):
    def test_correct_permissions(self):
        create_groups_permissions(
            {"group1": {"common.TenantSite": ["view", "add", "change", "delete"]}}
        )

    def test_add_permissions_to_existing_group(self):
        create_groups_permissions(
            {"group2": {"common.TenantSite": ["view", "add", "change"]}}
        )
        create_groups_permissions({"group2": {"common.TenantSite": ["delete"]}})

    def test_add_permissions_to_nonexisting_model(self):
        with self.assertRaises(LookupError):
            create_groups_permissions(
                {"group3": {"common.XYZ_does_not_exist": ["view", "add", "change"]}}
            )



"""
MedUX - A Free/OpenSource Electronic Medical Record
Copyright (C) 2017-2021 Christian González

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from medux.common.models import BaseModel
from .datapacks import PackageDataModel
from .organizations import HealthServiceProvider, Person
from django.db.models.enums import TextChoices

__all__ = [
    "Patient",
    "Encounter",
    "Problem",
    "NarrativeType",
    "Narrative",
]


class RelationType(models.Model):
    """A relation between 2 identities.

    This could be father, mother, son, daughter, uncle, friend etc."""

    name = models.CharField(max_length=255)


# class MaritalStatus(models.Model):
#     name = models.CharField(max_length=255)
#
#     class Meta:
#         verbose_name_plural = _("Marital statuses")
# TODO: maybe make marital status updateable


class MaritalStatusChoices(TextChoices):
    ANULLED = "A", _("Annulled")
    DIVORCED = "D", _("Divorced")
    INTERLOCUTORY = "I", _("Interlocutory")
    LEGALLY_SEPARATED = "L", _("Legally Separated")
    MARRIED = "M", _("Married")
    POLYGAMOUS = "P", _("Polygamous")
    NEVER_MARRIED = "S", _("Never Married")
    DOMESTIC_PARTNER = "T", _("Domestic partner")
    UNMARRIED = "U", _("Unmarried")
    WIDOWED = "W", _("Widowed")
    UNKNOWN = "UNK", _("Unknown")


class Patient(Person):
    # see https://www.hl7.org/fhir/valueset-marital-status.html#expansion
    # mandatory fields:
    # name(s), birth_date, gender
    person = models.OneToOneField(
        Person, parent_link=True, on_delete=models.PROTECT, related_name="+"
    )

    marital_status = models.CharField(
        max_length=3, null=True, blank=True, choices=MaritalStatusChoices.choices
    )

    general_practitioner = models.ForeignKey(
        "Physician",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        help_text=_("The patient's general practitioner"),
        related_name="+",
    )
    physicians = models.ManyToManyField(
        "Physician",
        blank=True,
        help_text=_("Physicians the patient consults regularly"),
        related_name="+",
    )

    related_persons = models.ManyToManyField(
        "Patient", through="Relationship", related_name="+"
    )

    # Link to another Person to be used as emergency contact
    emergency_contact = models.ManyToManyField(
        Person,
        help_text=_("Persons who should be contacted in case of an emergency"),
        related_name="+",
        blank=True,
    )

    # Free text emergency contact information.
    emergency_contact_freetext = models.CharField(max_length=255, blank=True, null=True)

    comment = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text=_("Generic comment for this patient, e.g. to identify him/her."),
    )

    # # TODO: implement generatePupic and set required=True
    # pupic = models.CharField(
    #     max_length=24,
    #     blank=True,
    #     null=True,
    #     editable=False,
    #     help_text=_(
    #         "Portable Unique Person "
    #         "Identification Code as per GNUmed "
    #         "white papers"
    #     ),
    # )
    #
    # def generate_pupic(self):
    #     """Generates a Portable Unique Person Identification according to
    #     GNUmed definition"""
    #     # TODO: implement PUPIC generation
    #     return ""


class Relationship(models.Model):
    """A related person model to be used as through= table."""

    from_patient_id = models.ForeignKey(
        Patient, on_delete=models.CASCADE, related_name="+"
    )
    to_patient_id = models.ForeignKey(
        Patient, on_delete=models.CASCADE, related_name="+"
    )
    relation_type = models.ForeignKey(RelationType, on_delete=models.PROTECT)


# TODO: use BaseModel?


class Encounter(BaseModel):
    """A clinical encounter between a person and the health care system."""

    date = models.DateTimeField(default=timezone.now)

    # FIXME: add a function to set this to "anonymous" in case of deletion of HSP
    provider = models.ForeignKey(HealthServiceProvider, on_delete=models.PROTECT)

    def __str__(self):
        return str(self.date)


class Problem(BaseModel):
    """Represents a clinical problem during a time.

    In other specifications, this could be named health issue, condition
    (FHIR), diagnosis etc.
    One or more (ICPC2/ICD10 coded) diagnoses could be a part of a problem,
    but there could be problems that have no clear diagnosis as well.
    """

    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class NarrativeType(models.Model):
    """The type of a narrative.

    This per default are the first 2 letters of narratives in the SOAP schema:
    S subjective, anamnesis
    O objective, findings
    A assessment, differential diagnoses
    P plan - what to do next
    """

    shortcut = models.CharField(primary_key=True, max_length=2)
    name = models.CharField(max_length=100)


class Narrative(BaseModel):
    type = models.ForeignKey(NarrativeType, on_delete=models.PROTECT)
    encounter = models.ForeignKey(Encounter, on_delete=models.CASCADE)

    problem = models.ForeignKey(Problem, on_delete=models.PROTECT)
    narrative = models.TextField()


class Firstname(PackageDataModel):
    """Model for saving packages with firstnames, matched with the correct gender.

    See https://github.com/MatthiasWinkelmann/firstname-database, under GNU Free Documentation License 1.2+
    """

    NAME_GENDER_CHOICES = (
        ("M", "Male"),
        ("1M", "male if first part of name, otherwise mostly female"),
        ("?M", "mostly male"),
        ("F", "Female"),
        ("1F", "female if first part of name, otherwise mostly male"),
        ("?F", "mostly female"),
        ("?", "unisex"),
    )
    name = models.CharField(max_length=50)
    sex = models.CharField(max_length=2, choices=NAME_GENDER_CHOICES)



from component_tags import template
from django.apps import apps

register = template.Library()


@register.tag()
class List(template.Component):
    """
    A flexible component to display a HTMX enabled list.
    """

    editable = template.Attribute(default=True)
    deletable = template.Attribute(default=True)
    model = template.Attribute(required=True)
    fields = []

    class Meta:
        template_name = "core/components/list.html"

    def get_queryset(self):
        app_label, model_name = self.attrs["model"].var.split(".")
        model = apps.get_model(app_label, model_name)
        return model.objects.all()

    def get_context_data(self, context):
        objects = self.get_queryset()
        context.update({"objects": objects})
        return super().get_context_data(context)



import pytest

from django.utils.timezone import datetime
from medux.utils import parse_date_string


# TODO test other localizations after implementing them - MMDDYY etc.
def test_parse_date_string_ddmmyyyy():
    assert parse_date_string("12111978") == datetime(1978, 11, 12).date()


def test_parse_date_string_ddmmyyyy_with_dots():
    assert parse_date_string("12.11.1978") == datetime(1978, 11, 12).date()


def test_parse_date_string_ddmmyy():
    assert parse_date_string("121178") == datetime(1978, 11, 12).date()


def test_parse_date_string_ddmmyy_with_dots():
    assert parse_date_string("12.11.78") == datetime(1978, 11, 12).date()


def test_parse_date_string_svnr():
    assert parse_date_string("1234150456") == datetime(1956, 4, 15).date()


def test_parse_date_string_wrong_format():
    assert parse_date_string("123456") == None



#  MedUX - Open Source Electronical Medical Record
#  Copyright (c) 2022  Christian González
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.views.generic import FormView
from medux.common.htmx.mixins import HtmxMixin

from medux.core.api import ICommand


class CommandLineView(HtmxMixin, FormView):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._command = None

    def execute(self):
        for plugin in ICommand:
            pass

    def parse(self):
        for plugin in ICommand:
            if self.command_text.startswith(plugin.shortcut):
                self._command = plugin



import logging

logger = logging.getLogger(__file__)


class DeviceMiddleware:
    """A middleware that adds the current device to the request.

    A device commonly is a computer in this context.
    Add ``medux.common.middleware.DeviceMiddleware`` to your MIDDLEWARE
    dict in settings.py.
    You can use the device in a template to your needs:

    .. code:: django

        <span>Device: {{ request.device }}</span>

    Or in a view:

    .. code:: python

        logger.debug(f"Request originating from device {request.device}.")
    """

    def __init__(self, get_response):
        """Initializes the middleware"""
        self.get_response = get_response

    def __call__(self, request):
        # TODO: implement device fetching in middleware
        request.device = None
        response = self.get_response(request)
        # logger.debug(f"Current device: {request.site.tenantsite}.")
        return response



#  MedUX - Open Source Electronical Medical Record
#  Copyright (c) 2022  Christian González
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.contrib.auth.models import Group
from django.db import models

from medux.common.models import User, Tenant
from medux.preferences.definitions import KeyType, Scope
from medux.preferences.models import ScopedPreference, BasePreference
from medux.preferences.registry import PreferencesRegistry
from medux.preferences.tools import get_effective_setting


class CachedPreferences:
    """Creates a request specific set of preferences, usable for templates.

    CachedPreferences allows easy access to the preferences for a specific context, given a request:

    CachedPreferences(request).my_namespace.my_setting returns the specific setting. It respects the current
    user, groups, device, and tenant

    # It creates a dict with nested preferences:
    # <namespace>
    # +-<key-1>
    #   +-user
    #   | +- value: 1
    #   | +- user: 23
    #   +-device
    #     +- value: 2
    #     +- device: 23543
    # ...
    """

    from django.http import HttpRequest

    class Namespace:
        """Transparently represents the first part (=namespace, before the dot) of a key, when
        accessed as CachedPreferences' attribute, e.g.
        ```python
        preferences = CachedPreferences(request)
        preferences.prescriptions.medication_message
        ```
        """

        from django.http import HttpRequest

        def __init__(self, request: HttpRequest, namespace: str):
            self.namespace = namespace
            self.request = request

        def __repr__(self):
            return f"<Namespace '{self.namespace}'>"

        def __getattr__(self, key: str) -> str | None:
            """get a list of preferences items with that key, ordered by scope:

            The scope order is: USER > DEVICE > GROUP > TENANT > VENDOR
            If a key is found during the traversal, its value is returned, else the next scope is
            looked up. If no key is found, None is returned.
            """

            setting = get_effective_setting(
                namespace=self.namespace,
                key=key,
                request=self.request,
            )
            # special case boolean.False, or None: return "" for  template variables
            if setting.key_type == KeyType.BOOLEAN and setting.value.lower() == "false":
                return ""
                # FIXME: this should be done better, distinguish between template and view code

            return setting.value

    def __init__(self, request: HttpRequest):

        self.request = request

    def __getattr__(self, namespace):
        """Helper to retrieve preferences key (all scopes) in a pythonic way:

        CachedPreferences.<namespace>.<key>"""

        return self.Namespace(self.request, namespace)

    @classmethod
    def get(
        cls,
        namespace: str,
        key: str,
        scope: Scope,
        full_object: bool = False,
        user: User | None = None,
        group: Group | None = None,
        device=None,  # TODO add device support
        tenant: Tenant | None = None,
    ) -> int | str | bool | models.Model | None:
        """Convenience method to retrieve a preferences key.

        :return: namespaced preferences value, according to given scope
            and, if applicable, the related object like user, group etc.
            If no object exists, return None.
            If pointing to an unregistered namespace/key, raise a
            KeyError.

        :param namespace: the namespace this key is saved under.
            Usually the app's name.
        :param key: the key to be retrieved
        :param scope: the scope that key is valid for. If scope is None,
            and more than one keys are saved under that scope, the key
            with the highest priority is taken:
            USER > DEVICE > GROUP > TENANT > VENDOR
        :param key_type: preferences type: "str", "bool", "int"
        :param full_object: if True, return the complete model instance
            instead of only its value
        :param user: if scope is USER, you have to provide a User object
            that key/scope is valid for.
        :param group: if scope is GROUP, you have to provide a
            SettingsGroup object that key/scope is valid for.
        :param device: if scope is DEVICE, you have to provide a Device
            object that key/scope is valid for.
        :param tenant: if scope is TENANT, you have to provide a Tenant
            object that key/scope is valid for.
        """

        ScopedPreference.assure_exists(namespace, key, scope)
        key_type = PreferencesRegistry.key_type(namespace, key)

        filters = {
            "base__namespace": namespace,
            "base__key": key,
        }
        if scope:
            filters["scope"] = scope
            if scope == Scope.USER:
                filters["user"] = user
            elif scope == Scope.GROUP:
                filters["group"] = group
            elif scope == Scope.DEVICE:
                filters["device"] = device
            elif scope == Scope.TENANT:
                filters["tenant"] = tenant

        objects = ScopedPreference.objects.filter(**filters)
        if len(objects) == 0:
            # if this setting does not exist (yet), return None
            return None

        if len(objects) == 1:
            obj = objects.first()
            if full_object:
                return obj
            value = obj.value  # type: str
        else:
            # more than one scopes under that key and scope
            ids = ",".join(i.id for i in objects)
            raise KeyError(
                f"There are multiple preferences keys under {namespace}.{key}[{scope}]: ids {ids}"
            )
        # filter out int and boolean values and return them instead.
        if key_type == KeyType.INTEGER:
            return int(value)
        if key_type == KeyType.BOOLEAN:
            if value.lower() == "false":
                return False
            elif value.lower() == "true":
                return True
            else:
                raise ValueError(f"DB value '{value}' cannot be casted into boolean!")
        if key_type in (KeyType.STRING, KeyType.TEXT):
            return str(value)

        # should never happen...
        raise KeyError(f"Unknown preferences type: {key_type}")

    @classmethod
    def set(
        cls,
        namespace: str,
        key: str,
        value: str | int | bool,
        scope: Scope,
        user=None,
        tenant=None,
        device=None,
        group: Group = None,
    ) -> None:

        """Convenience method to set a preferences key.

        Raises a TypeError if value is of incorrect KeyType
        """

        ScopedPreference.assure_exists(namespace, key, scope)

        # key_type is not used directly, as the value in the DB always is casted to a str.
        # Retrieving later is casted by the type saved in the PreferencesRegistry.

        key_type = PreferencesRegistry.key_type(namespace, key)
        if key_type == KeyType.INTEGER:
            # try to cast value into an integer. If not possible,
            # an Exception is raised.
            int(value)
        elif key_type == KeyType.BOOLEAN:
            if value not in (True, False):
                raise TypeError("value must be a Boolean.")

        filter = {
            "scope": scope,
        }
        # set correct scope
        if scope == Scope.USER:
            if user is None:
                raise AttributeError(
                    "When scope==USER, a user object must be provided."
                )
            filter["user"] = user

        elif scope == Scope.DEVICE:
            if device is None:
                raise AttributeError(
                    "When scope==DEVICE, a device object must be provided."
                )
            filter["device"] = device

        elif scope == Scope.GROUP:
            if group is None:
                raise AttributeError(
                    "When scope==GROUP, a group object must be provided."
                )
            filter["group"] = group

        elif scope == Scope.TENANT:
            if tenant is None:
                raise AttributeError(
                    "When scope==TENANT, a tenant object must be provided."
                )
            filter["tenant"] = tenant

        # TODO: maybe do some casting / checks here?

        # if base setting does not exist, create it
        base, base_created = BasePreference.objects.get_or_create(
            namespace=namespace, key=key
        )
        if base_created:
            base.save()
        filter["base"] = base

        # get_or_create can't be used, as old / new value could differ,
        # so get() must not include it.
        try:
            item = ScopedPreference.objects.get(**filter)
        except ScopedPreference.DoesNotExist:
            filter["value"] = value
            item = ScopedPreference.objects.create(**filter)

        item.value = str(value)
        item.save()


# @receiver(request_finished)
# def on_request_finished(sender, **kwargs):
#     preferences.invalidate()



# Generated by Django 4.1 on 2022-10-17 07:45

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import medux.preferences.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("common", "0002_create_fixtures"),
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BasePreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "namespace",
                    models.CharField(
                        max_length=25,
                        validators=[medux.preferences.models.validate_lower],
                    ),
                ),
                (
                    "key",
                    models.CharField(
                        max_length=255,
                        validators=[medux.preferences.models.validate_lower],
                    ),
                ),
            ],
            options={
                "verbose_name": "Settings",
                "verbose_name_plural": "Settings",
                "ordering": ["namespace", "key"],
            },
        ),
        migrations.CreateModel(
            name="ScopedPreference",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "scope",
                    models.IntegerField(
                        choices=[
                            (1, "Vendor"),
                            (2, "Tenant"),
                            (3, "Group"),
                            (4, "Device"),
                            (5, "User"),
                        ]
                    ),
                ),
                ("value", models.CharField(max_length=255, null=True)),
                (
                    "base",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="scopedpreference",
                        to="preferences.basepreference",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="auth.group",
                    ),
                ),
                (
                    "tenant",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="common.tenant",
                        verbose_name="Tenant",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Scoped preferences",
                "verbose_name_plural": "Scoped preferences",
                "ordering": ["base", "scope"],
                "permissions": [
                    (
                        "change_own_user_scopedpreference",
                        "Can change own user's preferences",
                    ),
                    (
                        "change_own_tenant_scopedpreference",
                        "Can change own tenant's preferences",
                    ),
                    (
                        "change_own_group_scopedpreference",
                        "Can change own groups' preferences",
                    ),
                    (
                        "change_device_scopedpreference",
                        "Can change a device's preferences",
                    ),
                ],
                "unique_together": {
                    ("base", "scope", "user"),
                    ("base", "scope", "group"),
                    ("base", "scope", "tenant"),
                },
            },
        ),
    ]



from enum import Enum

from django.db import models
from django.utils.translation import gettext_lazy as _

from ..fields import CodeField, PeriodField, QuantityField
from medux.common.models import BaseModel


# https://hl7.org/fhir/observation.html#Observation
class Observation(BaseModel):
    """Measurements and simple assertions made about a patient,
    device or other subject.

    .. note:

        IMPORTANT: inheriting classes must implement at least the "value" field for reading!
        This could be done as simple CharField, or as a property method that gets the value
        from more specific fields.

    Use cases:

    * Vital signs such as body weight, blood pressure, and temperature
    * Laboratory Data like blood glucose, or an estimated GFR
    * Imaging results like bone density or fetal measurements
    * Clinical Findings such as abdominal tenderness
    * Device measurements such as EKG data or Pulse Oximetry data
    * Clinical assessment tools such as APGAR or a Glasgow Coma Score
    * Personal characteristics: such as eye-color
    * Social history like tobacco use, family support, or cognitive status
    * Core characteristics like pregnancy status, or a death assertion
    """

    class Status(Enum):
        registered = "registered"
        preliminary = "preliminary"
        final = "final"
        amended = "amended"

    # The status of the result value
    # http://hl7.org/fhir/ValueSet/observation-status
    status = CodeField(terminology_binding=Status)

    # The time or time-period the observed value is asserted as being true. For biological subjects -
    # e.g. human patients - this is usually called the "physiologically relevant time".
    # This is usually either the time of the procedure or of specimen collection, but very often the source
    # of the date/time is not known, only the date/time itself
    # Type: dateTime|Period|Timing

    effective_datetime = models.DateTimeField(
        blank=True,
        null=True,
        help_text=_("Clinically relevant time for observation"),
    )

    # TODO: add keyword arguments
    effective_period = PeriodField(
        # blank=True,
        # help_text=_("Clinically relevant time-period for observation"),
    )

    encounter = models.ForeignKey(
        "Encounter",
        null=True,
        # blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    # The date and time this version of the observation was made available to
    # providers, typically after the results have been reviewed and verified
    issued = models.DateTimeField(blank=True, null=True)

    # Who was responsible for asserting the observed value as "true"
    # FIXME: + Patient, Person...
    performer = models.ForeignKey(
        "HealthServiceProvider",
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    # The information determined as a result of making the observation,
    # if the information has a simple value
    # Type: Quantity|str|bool|int|Range|Ratio|time|dateTime|Period
    # IMPORTANT: has to be implemented by inheriting Model classes
    # value = ...

    # Provides a reason why the expected value in the element Observation.value[x] is missing
    # dataAbsentReason = CodeableConceptField(
    #     "DataAbsentReason", null=True, related_name="+"
    # )

    # interpretation = CodeableConceptField(
    #     "Observation Interpretation", null=True, related_name="+"
    # )

    comment = models.CharField(max_length=255, blank=True)

    # see http://hl7.org/fhir/ValueSet/body-site
    # bodySite = CodeableConceptField("SNOMED CT Body Structures??", null=True)

    # method = models.ForeignKey("ObservationMethods", null=True, related_name="+")

    # specimen = models.ForeignKey(
    #     "Specimen", null=True, on_delete=models.SET_NULL, related_name="+"
    # )

    device = models.ForeignKey(
        "Device",
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    # derivedFrom = ManyReferenceField(
    #     "DocumentReference|ImagingStudy|"
    #     "Media|QuestionnaireResponse|"
    #     "Observation|Sequence"
    # )

    # component = models.ManyToManyField(Component)

    # referenceRange = models.ManyToManyField(ReferenceRange)

    # whether this observation is a group observation (e.g. a battery, a panel of
    # tests, a set of vital sign measurements) that includes the target as a
    # member of the group.
    # When using this element, an observation will typically have either a
    # value or a set of related resources, although both may be present in
    # some cases. For a discussion on the ways Observations can assembled in
    # groups together.
    def has_member(self):
        return False

    def __str__(self):
        if hasattr(self, "value"):
            return f"{self.velue}"
        else:
            return super().__str__()


class PatientObservation(Observation):
    """Observation that is taken from a patient."""

    subject = models.ForeignKey(
        "Patient",
        on_delete=models.CASCADE,
        related_name="+",
    )

    class Meta:
        abstract = True


class DeviceObservation(Observation):
    """Observation that is taken from a device."""

    subject = models.ForeignKey(
        "Device",
        on_delete=models.CASCADE,
        related_name="+",
    )

    class Meta:
        abstract = True


# class BloodPressure(PatientObservation):
#     value=RatioField()


class BloodPressure(PatientObservation):
    systolic = models.PositiveSmallIntegerField()
    diastolic = models.PositiveSmallIntegerField()

    def value(self):
        return f"{self.systolic}/{self.diastolic}"


class BodyWeight(PatientObservation):
    value = QuantityField(max_digits=8, decimal_places=2)


class BodyTemperature(PatientObservation):
    value = QuantityField(max_digits=3, decimal_places=1)



import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from medux.preferences.cached_preferences import CachedPreferences
from medux.preferences.definitions import Scope, KeyType
from medux.preferences.models import ScopedPreference
from medux.preferences.registry import PreferencesRegistry


@pytest.fixture
def user_admin():
    admin = get_user_model().objects.create(username="admin")
    admin.save()
    return admin


@pytest.fixture
def user_nobody():
    nobody = get_user_model().objects.create(username="nobody")
    nobody.save()
    return nobody


@pytest.fixture
def dummy_group():
    group = Group.objects.create(name="dummy")
    group.save()
    return group


@pytest.mark.django_db
def test_preferences_get_nonexisting():
    """test if non-existing setting returns None"""
    PreferencesRegistry.register("test", "non_existing", [Scope.VENDOR])
    result = CachedPreferences.get("test", "non_existing", Scope.VENDOR)
    assert result is None


@pytest.mark.django_db
def test_preferences_integer(user_admin, user_nobody):
    """test if an int preferences returns actually an int"""
    PreferencesRegistry.register(
        "test", "int_setting", [Scope.USER], key_type=KeyType.INTEGER
    )
    ScopedPreference.set(
        "test",
        "int_preferences",
        42,
        Scope.USER,
        user=user_admin,
    )

    result = CachedPreferences.get(
        "test", "int_preferences", Scope.USER, user=user_admin
    )
    assert result == 42
    assert type(result) == int

    result = CachedPreferences.get(
        "test", "int_preferences", Scope.USER, user=user_nobody
    )
    assert result is None


@pytest.mark.django_db
def test_spreferences_integer_casting(admin_user):
    """test if an int preferences returns actually an int"""
    PreferencesRegistry.register(
        "test", "cast_int_preferences", [Scope.USER], key_type=KeyType.INTEGER
    )
    ScopedPreference.set(
        "test", "cast_int_preferences", "42", Scope.USER, user=admin_user
    )

    assert (
        CachedPreferences.get(
            "test", "cast_int_preferences", Scope.USER, user=admin_user
        )
        == 42
    )


@pytest.mark.django_db
def test_preferences_str(admin_user):
    """test if a str preferences actually returns a str"""
    PreferencesRegistry.register(
        "test", "str_setting", [Scope.USER], key_type=KeyType.STRING
    )
    ScopedPreference.set("test", "str_preferences", "42a", Scope.USER, user=admin_user)

    assert (
        CachedPreferences.get("test", "str_preferences", Scope.USER, user=admin_user)
        == "42a"
    )


@pytest.mark.django_db
def test_preferences_bool_true(admin_user):
    """test if a bool:True preferences actually returns bool:True"""
    PreferencesRegistry.register(
        "test", "true_preferences", [Scope.USER], key_type=KeyType.BOOLEAN
    )
    ScopedPreference.set("test", "true_preferences", True, Scope.USER, user=admin_user)
    result = CachedPreferences.get(
        "test", "true_preferences", Scope.USER, user=admin_user
    )
    assert result is True


@pytest.mark.django_db
def test_preferences_bool_false(admin_user):
    """test if a bool:False preferences actually returns bool:False"""
    PreferencesRegistry.register(
        "test", "false_preferences", [Scope.USER], key_type=KeyType.BOOLEAN
    )
    ScopedPreference.set(
        "test", "false_preferences", False, Scope.USER, user=admin_user
    )
    result = CachedPreferences.get(
        "test", "false_preferences", Scope.USER, user=admin_user
    )
    assert result is False


@pytest.mark.django_db
def test_spreferences_set_get_USER(admin_user, user_nobody):
    """test if a USER scoped setting returns the correct user and key"""
    PreferencesRegistry.register(
        "test", "user_preferences", [Scope.USER], key_type=KeyType.STRING
    )
    ScopedPreference.set("test", "user_preferences", "foo", Scope.USER, user=admin_user)

    assert (
        CachedPreferences.get("test", "user_preferences", Scope.USER, user=admin_user)
        == "foo"
    )
    assert (
        CachedPreferences.get("test", "user_preferences", Scope.USER, user=user_nobody)
        is None
    )


@pytest.mark.django_db
def test_preferences_set_get_USER_overwrite(admin_user):
    """test if a USER scoped setting returns the correct user and key"""
    PreferencesRegistry.register(
        "test", "user_preferences", [Scope.USER], key_type=KeyType.STRING
    )
    ScopedPreference.set("test", "user_preferences", "foo", Scope.USER, user=admin_user)
    ScopedPreference.set(
        "test", "user_preferences", "foo2", Scope.USER, user=admin_user
    )

    assert (
        CachedPreferences.get("test", "user_preferences", Scope.USER, user=admin_user)
        == "foo2"
    )


@pytest.mark.django_db
def test_preferences_set_get_VENDOR():
    PreferencesRegistry.register(
        "test", "vendor_preferences", [Scope.VENDOR], key_type=KeyType.STRING
    )
    ScopedPreference.set("test", "vendor_preferences", "shdkjhf", Scope.VENDOR)

    assert (
        CachedPreferences.get("test", "vendor_preferences", Scope.VENDOR) == "shdkjhf"
    )


@pytest.mark.django_db
def test_preferences_set_get_VENDOR_auto_keytype():
    PreferencesRegistry.register("test", "vendor_preferences", [Scope.VENDOR])
    ScopedPreference.set("test", "vendor_preferences", "shdkjhf", Scope.VENDOR)

    assert (
        CachedPreferences.get("test", "vendor_preferences", Scope.VENDOR) == "shdkjhf"
    )


@pytest.mark.django_db
def test_preferencesset_get_default_scope():
    PreferencesRegistry.register(
        "test", "vendor_preference2", [Scope.VENDOR], KeyType.INTEGER
    )
    ScopedPreference.set("test", "vendor_preference2", 42, Scope.VENDOR)

    result = CachedPreferences.get("test", "vendor_preference2", Scope.VENDOR)
    assert result == 42
    pass


@pytest.mark.django_db
def test_preferences_set_get_without_user_fk_object():
    PreferencesRegistry.register(
        "test", "vendor_preferences", [Scope.USER], key_type=KeyType.INTEGER
    )
    with pytest.raises(AttributeError):
        ScopedPreference.set("test", "vendor_preferences", 45, Scope.USER)


@pytest.mark.django_db
def test_preferences_set_get_specific_scope(user_admin, user_nobody, dummy_group):
    """Make sure scope with the highest priority is returned when more than one scopes
    are available for a key"""
    PreferencesRegistry.register(
        "test",
        "multiple_preferences",
        [Scope.VENDOR, Scope.USER, Scope.GROUP],
        key_type=KeyType.INTEGER,
    )
    ScopedPreference.set("test", "multiple_preferences", 42, Scope.VENDOR)
    ScopedPreference.set(
        "test", "multiple_preferences", 43, Scope.USER, user=user_admin
    )
    ScopedPreference.set(
        "test", "multiple_preferences", 52, Scope.GROUP, group=dummy_group
    )  # dummy

    assert (
        CachedPreferences.get(
            "test", "multiple_preferences", Scope.USER, user=user_admin
        )
        == 43
    )



from django.conf import settings
from django.conf.urls.static import static
from django.contrib.auth.views import LogoutView
from django.urls import path

from medux.common.views import LoginView

# namespaced URLs
app_name = "common"

# URLs namespaced  under common/
urlpatterns = [
    # path("", views.IndexView.as_view(), name="index"),
]


# global URLs
root_urlpatterns = [
    path(
        "accounts/login/",
        LoginView.as_view(),
        name="login",
    ),
    path(
        "accounts/logout/",
        LogoutView.as_view(template_name="registration/logout.html"),
        name="logout",
    ),
]

if settings.DEBUG:
    root_urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from crispy_forms.layout import Div
from crispy_forms.utils import flatatt


class Card(Div):
    """
    Layout object. It wraps fields in a Card.

    Example:
        Card(
            'form_field_1',
            'form_field_2',
            title="Sample card"
            subtitle="Always play nice."
        )
    """

    template = "%s/layout/card.html"

    def __init__(
        self,
        *fields,
        title="",
        title_id="",
        title_class="",
        subtitle="",
        **kwargs,
    ):

        super().__init__(*fields, **kwargs)
        self.title = title
        self.title_id = title_id or f"card-title-{hash(self)}"
        self.title_class = title_class
        self.subtitle = subtitle

        kwargs = {
            **kwargs,
            "aria-labelledby": "%s-label" % self.title_id,
        }

        self.flat_attrs = flatatt(kwargs)



import re
from typing import Union
from datetime import datetime
from dateutil.relativedelta import relativedelta
from django.utils.datetime_safe import date
from django.utils import formats

# get localized date input formats
input_formats = formats.get_format_lazy("DATE_INPUT_FORMATS")


def parse_date_string(date_string: str) -> Union[date, None]:
    """Tries to guess which date format the date_string has and returns a correct date object."""
    if not date_string:
        return None
    date_string = date_string.strip()
    if not re.sub("/|\.", "", date_string).isnumeric():
        return None
    # this does not work. Input formats are in an order we cant anticipate, and this could create
    # situations where a DE localizeable fast-entered "131279" is not parsed into datetime(1979, 12, 13), but
    # the first format in DE is used (%d.%m.%Y) which tries to interpret the last 4 places as YEAR, so it
    # parses datetime(1279, 3, 1) out of it...
    # for input_format in input_formats:  # type: str
    #     stripped_format = re.sub(",|\.|/| ", "", input_format)
    #     if search_date := datetime.strptime(date_string, stripped_format):
    #         return search_date.date()
    # return None
    if len(date_string) == 6:
        # DDMMYY
        if re.match(r"^[0123]\d[01]\d\d\d$", date_string):
            return datetime.strptime(date_string, "%d%m%y").date()
    elif len(date_string) == 8:
        # DD.MM.YY
        if re.match(r"^[0123]\d\.[01]\d\.\d\d$", date_string):
            return datetime.strptime(date_string, "%d.%m.%y").date()
        # DDMMYYYY
        if re.match(r"^[0123]\d[01]\d[12][90]\d\d$", date_string):
            return datetime.strptime(date_string, "%d%m%Y").date()
    elif len(date_string) == 10:
        # DD.MM.YYYY
        if re.match(r"^[0123]\d\.[01]\d\.[12][90]\d\d$", date_string):
            return datetime.strptime(date_string, "%d.%m.%Y").date()
        elif date_string.isnumeric():  # SVNR + DDMMYY
            dt = datetime.strptime(date_string[4:], "%d%m%y").date()
            if dt > datetime.now().date():
                dt -= relativedelta(years=100)
            return dt
    return None



from django.contrib import admin

from .models import Vendor


@admin.register(Vendor)
class VendorAdmin(admin.ModelAdmin):
    def has_add_permission(self, request):
        return super().has_add_permission(request) and not Vendor.objects.exists()



"""medux URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
import logging
from django.urls import path, include
from django.contrib import admin
from django.conf.urls.static import static
from django.conf import settings
from gdaps.pluginmanager import PluginManager


logger = logging.getLogger(__name__)

app_name = "medux"

PluginManager.load_plugin_submodule("urls")

urlpatterns = PluginManager.urlpatterns() + [
    path("admin/", admin.site.urls),
]

if settings.DEBUG_TOOLBAR:
    try:
        import debug_toolbar

        urlpatterns += (path("__debug__/", include(debug_toolbar.urls)),)
        urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
    except ImportError:
        logger.warning(
            "Debug Toolbar not available. Please install it or set DEBUG to False."
        )
# from django.contrib.staticfiles.urls import staticfiles_urlpatterns
# from django.contrib.staticfiles import views
#
# views.serve()



import logging

import toml
from django.db import transaction

from medux.preferences.cached_preferences import CachedPreferences
from medux.preferences.definitions import Scope
from medux.preferences.models import ScopedPreference
from medux.preferences.registry import PreferencesRegistry

logger = logging.getLogger(__file__)


class PreferencesLoader:
    """Base for Preferences loaders.

    You have to provide a Scope where this preferences go into (defaults to VENDOR)
    """

    def __init__(self, scope: Scope = Scope.VENDOR, foreign_object=None):
        # FIXME: change foreign_object to user/tenant/etc.?
        self.scope = scope
        if scope == scope.VENDOR:
            self.foreign_object = {}
        else:
            self.foreign_object = {scope.name.lower(): foreign_object}

    def load(self):
        """load the preferences file"""
        raise NotImplementedError

    def save(self):
        raise NotImplementedError

    def update_preferences(self, dct):
        """Updates the preferences using a dict:
        {
            "namespace_a":
                "let_user_decide": True,
                "other_preference": 42,
            },
            "namespace_b": {...}
        }
        """
        # wrong_keys = []
        with transaction.atomic():
            for namespace, keys in dct.items():
                for key in keys:
                    # allow not settin that is not registered
                    if not PreferencesRegistry.exists(namespace, key, self.scope):
                        raise KeyError(
                            f"Key {namespace}.{key} is not registered and can't be imported. Please register first."
                        )
                    CachedPreferences.set(
                        namespace=namespace,
                        key=key,
                        scope=self.scope,
                        value=dct[namespace][key],
                        **self.foreign_object,
                    )  # type: ScopedPreference


class TomlPreferencesLoader(PreferencesLoader):
    def __init__(self, filename, scope: Scope = Scope.VENDOR, foreign_object=None):
        super().__init__(scope, foreign_object)
        self._filename = filename

    def load(self):
        try:
            with open(self._filename, "r") as f:
                self.update_preferences(toml.load(f))
        except FileNotFoundError:
            raise
        except toml.TomlDecodeError as e:
            logger.error(f"Could load load file {self._filename}: {e}")


# TODO make non-VENDOR bulk-updates work



from django.db import models
from django.db.models import PositiveIntegerField
from django.utils.timezone import now
from django.utils.translation import gettext as _, pgettext

from .datapacks import PackageDataModel
from medux.common.models import BaseModel
from medux.core.fields import PeriodField


def ongoing_period() -> str:
    return ";".join([now().isoformat(), ""])


class Element(BaseModel):
    """The base definition for all elements contained inside a resource.

    All elements, whether defined as a Data Type (including primitives) or as
    part of a resource structure, have this base content:

    * Extensions
    * an internal id
    """

    # This field originally is named "extension". This causes some problems
    # with name clashing. So, as it's a ManyToManyField anyway, we renamed
    # it to "extensions"
    # extensions = models.ManyToManyField("Extension", related_name="+", blank=True)

    class Meta:
        abstract = True

    def to_xml(self):
        raise NotImplementedError

    def to_json(self):
        raise NotImplementedError


# class Extension(Element):
#     # SHALL be a URL, not a URN (e.g. not an OID or a UUID),
#     url = UriField()
#
#     # FIXME: this field in reality should be more flexible.
#     # see http://build.fhir.org/extensibility.html#Extension
#     # value = models.CharField(max_length=255, blank=True)
#
#     class Meta:
#         abstract = True


#
# class Period(models.Model):
#     """A time period defined by a start and end date/time.
#
#     If the start element is missing, the start of the period is not known.
#     If the end element is missing, it means that the period is ongoing, or the
#     start may be in the past, and the end date in the future, which means that
#     period is expected/planned to end at the specified time.
#     """
#
#     start = OptionalTimeDateTimeField(blank=True, null=True)
#     end = OptionalTimeDateTimeField(blank=True, null=True)
#
#     def __str__(self):
#         # TODO: return date with local format
#         return f"{self.start or _('unknown')} - {self.end or _('ongoing')}"
#
#     class Meta:
#         pass
#         # constraints = [
#         #     # Check if either start or end is presents
#         #     models.CheckConstraint(
#         #         check=models.Q(start__isnull=False) | models.Q(end__isnull=False),
#         #         name="either_start_or_end_present",
#         #     )
#         # ]


class ContactPoint(Element):
    """Details for all kinds of technology-mediated contact points.

    These can be for a person or organization, and includes telephone, email, etc.
    """

    CONTACT_POINT_SYSTEM = (
        ("phone", _("Phone")),
        ("fax", _("Fax")),
        ("email", _("Email")),
        ("pager", _("Pager")),
        ("url", _("URL")),
        ("sms", _("SMS")),
        ("other", _("Other")),
    )

    CONTACT_POINT_USE = (
        (
            "home",
            pgettext("at home", "Home"),
        ),
        ("work", _("Work")),
        ("temp", _("Temporary")),
        ("old", _("Old")),  # not in use anymore, or was never correct
        ("mobile", _("Mobile")),
    )

    # links to ContactPointSystem
    # http://hl7.org/fhir/ValueSet/contact-point-system
    system = models.CharField(max_length=20, choices=CONTACT_POINT_SYSTEM)

    value = models.CharField(max_length=255, blank=True)

    # links to ContactPointUse
    # http://hl7.org/fhir/ValueSet/contact-point-use
    use = models.CharField(max_length=20, choices=CONTACT_POINT_USE, blank=True)

    # in FHIR, this is actually "rank" - but could be easily converted when exporting.
    # we stay with "weight" to be consistent with MedUX
    weight = PositiveIntegerField(blank=True, default=1)

    # TODO: default=ongoing_period, add help_text, see bug
    # https://github.com/bikeshedder/django-composite-field/issues/5
    period = PeriodField(
        # help_text=_("The period in which this contact point is valid.")
    )

    def __str__(self):
        return f"{[value for key, value in self.CONTACT_POINT_SYSTEM if key == self.system][0]}: {self.value}"


# https://fhir-ru.github.io/codesystem-administrative-gender.html
class AdministrativeGender(models.Model):
    #: the (mostly single char) gender, like "m", "f", etc.
    tag = models.CharField(primary_key=True, max_length=2)

    label = models.CharField(max_length=50)
    sort_weight = models.IntegerField(default=0)
    comment = models.CharField(max_length=255, blank=True)

    def __str__(self):
        return self.label


# http://hl7.org/fhir/CodeSystem/medication-statement-status
class CodeValueSet(PackageDataModel):
    """Represents a ValueSet for Status items, like EpisodeStatus, MedicationStatus.

    You have to subclass this model to make use of it, as it is an abstract model.

    Each instance has the minimum fields:
    * code: the code that represents this item
    * display: the human readable code in short form
    * definition: a detailed description of the item
    """

    code = models.CharField(max_length=50)
    display = models.CharField(max_length=50)
    definition = models.CharField(max_length=255)

    class Meta:
        abstract = True

    def __str__(self):
        return self.code


# http://hl7.org/fhir/valueset-medication-status.html


# https://www.hl7.org/fhir/medication.html


# https://www.hl7.org/fhir/medication.html


# https://hl7.org/fhir/episode-of-care-status
class EpisodeStatus(CodeValueSet):
    pass

    class Meta:
        verbose_name_plural = _("Episode Statuses")


class EpisodeStatusHistory(models.Model):
    """ManyToMany "Through" model for EpisodeOfCare"""

    status = models.ForeignKey(EpisodeStatus, on_delete=models.CASCADE)
    episode = models.ForeignKey("EpisodeOfCare", on_delete=models.CASCADE)

    # TODO: add default=ongoing_period
    period = PeriodField()
    # period = models.ForeignKey(Period, on_delete=models.PROTECT, related_name="+")

    class Meta:
        verbose_name_plural = _("Episode Status Histories")


class EpisodeOfCare(BaseModel):
    """An episode of problem(s) within a given period."""

    ##### TODO

    # code: planned | waitlist | active | onhold | finished | cancelled | entered-in-error
    status = models.ForeignKey(
        EpisodeStatus, on_delete=models.PROTECT, related_name="+"
    )
    status_history = models.ManyToManyField(EpisodeStatus, through=EpisodeStatusHistory)



from io import BytesIO
from logging import getLogger
from pathlib import Path

from django.core.files import File
from django.db.models import Model  # FIXME could be a problem during migrations
from PIL import Image

logger = getLogger(__file__)


def create_groups_permissions(
    groups_permissions: dict[str, dict[Model | str, list[str]]]
):
    """Creates groups and their permissions defined in given `groups_permissions` automatically.

    :param groups_permissions: a dict, see also `MeduxPluginAppConfig.groups_permissions`

    Based upon the work here: https://newbedev.com/programmatically-create-a-django-group-with-permissions

    """
    from django.apps import apps
    from django.contrib.auth.models import Group, Permission
    from django.contrib.contenttypes.models import ContentType

    for group_name in groups_permissions:

        # Get or create group (even if there are no permissions
        # to save in the dict)
        group, created = Group.objects.get_or_create(name=group_name)
        if created:
            group.save()

        # Loop models in group
        for model in groups_permissions[group_name]:
            # if model_class is written as dotted str, convert it to class
            if type(model) is str:
                model_class = apps.get_model(model)
            else:
                model_class = model

            # Loop permissions in group/model

            for perm_name in groups_permissions[group_name][model]:

                # Generate permission name as Django would generate it
                codename = f"{perm_name}_{model_class._meta.model_name}"

                try:
                    # Find permission object and add to group
                    content_type = ContentType.objects.get(
                        app_label=model_class._meta.app_label,
                        model=model_class._meta.model_name.lower(),
                    )
                    perm = Permission.objects.get(
                        content_type=content_type,
                        codename=codename,
                    )
                    group.permissions.add(perm)
                    logger.info(
                        f"  Adding permission '{codename}' to group '{group.name}'"
                    )
                except Permission.DoesNotExist:
                    logger.critical(f"  ERROR: Permission '{codename}' not found.")


def str_to_bool(bool_str: str) -> bool:
    """:return a bool True if bool_str is "true", else False."""
    return bool_str.lower() == "true"


def snake_case_to_spaces(string):
    return string.replace("_", " ")


image_types = {
    "jpg": "JPEG",
    "jpeg": "JPEG",
    "png": "PNG",
    "gif": "GIF",
    "tif": "TIFF",
    "tiff": "TIFF",
}


def resize_image(image, width, height):
    """Resizes an image to a given width/height"""
    # thanks to https://stackoverflow.com/a/71919278/818131

    img = Image.open(image)
    if img.width > width or img.height > height:
        output_size = (width, height)
        img.thumbnail(output_size)
        img_filename = Path(image.file.name).name
        # get file extension
        img_suffix = Path(image.file.name).name.split(".")[-1]
        # Use file extension to determine  file type from the image_type's dictionary
        img_format = image_types[img_suffix]
        # Save resized image into the buffer with the correct file type
        buffer = BytesIO()
        img.save(buffer, format=img_format)
        # Wrap buffer in a File object
        file_object = File(buffer)
        # Save new resized file
        image.save(img_filename, file_object)



from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _

from . import __version__
from .api import MeduxPluginAppConfig


class CommonConfig(MeduxPluginAppConfig):
    """A GDAPS Django app plugin.

    It needs a special parameter named ``PluginMeta``. It is the key for GDAPS
    to recognize this app as a GDAPS plugin.
    ``PluginMeta`` must point to a class that implements certain attributes
    and methods.
    """

    name = "medux.common"
    default = True  # FIXME: Remove when django bug is fixed
    # greate empty users group
    groups_permissions = {"Users": {}}

    class PluginMeta:
        """This configuration is the introspection data for plugins."""

        # the plugin machine "name" is taken from the AppConfig, so no name here
        verbose_name = _("Common")
        author = "Christian González"
        author_email = "office@nerdocs.at"
        vendor = "Nerdocs"
        description = _(
            "MedUX common tools and models, which are used for medux_online and medux"
        )
        category = _("Base")
        visible = True
        version = __version__
        # compatibility = "medux.core>=2.3.0"

    def initialize(self):
        pass

    def ready(self):
        from . import signals

        post_save.connect(signals.add_user_to_users_group, sender=get_user_model())



"""
MedUX - A Free/OpenSource Electronic Medical Record
Copyright (C) 2017 Christian González

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from rest_framework import serializers
from medux.core.models import Patient, Identity, Name, Address


class IdentitySerializer(serializers.ModelSerializer):
    class Meta:
        model = Identity
        fields = [
            "pupic",
            "gender",
            "dob",
            "dob_is_estimated",
            "marital_status",
            "cob",
            "deceased",
            "title",
            "emergency_contact_freetext",
            "emergency_contact",
            "comment",
        ]


class NameSerializer(serializers.ModelSerializer):
    """
    {
        "use" : "<code>", // usual | official | temp | nickname | anonymous | old | maiden
        "text" : "<string>", // Text representation of the full name
        "family" : "<string>", // Family name (often called 'Surname')
        "given" : ["<string>"], // Given names (not always 'first'). Includes middle names
        "prefix" : ["<string>"], // Parts that come before the name
        "suffix" : ["<string>"], // Parts that come after the name
        "period" : { Period } // Time period when name was/is in use
    }
    """

    use = serializers.SerializerMethodField()
    text = serializers.SerializerMethodField()
    family = serializers.SerializerMethodField()
    given = serializers.SerializerMethodField()
    prefix = serializers.SerializerMethodField()
    suffix = serializers.SerializerMethodField()
    period = serializers.SerializerMethodField()

    def get_use(self, name):
        return "usual"  # Todo: This information is not beeing stored in the DB

    def get_text(self, name):
        return "{} {}".format(name.firstname, name.lastname)

    def get_family(self, name):
        return name.lastname

    def get_given(self, name):
        return name.firstname

    def get_prefix(self, name):
        return ""  # Todo: This information is not beeing stored in the DB

    def get_suffix(self, name):
        return ""  # Todo: This information is not beeing stored in the DB

    def get_period(self, name):
        return ""  # Todo: This information is not beeing stored in the DB

    class Meta:
        model = Name
        fields = [
            "use",
            "text",
            "family",
            "given",
            "prefix",
            "suffix",
            "period",
        ]


class ContactPointSerializer(serializers.ModelSerializer):
    """
    See http://build.fhir.org/datatypes.html#ContactPoint
    {
      "system" : "<code>", // C? phone | fax | email | pager | url | sms | other
      "value" : "<string>", // The actual contact point details
      "use" : "<code>", // home | work | temp | old | mobile - purpose of this contact point
      "rank" : "<positiveInt>", // Specify preferred order of use (1 = highest)
      "period" : { Period } // Time period when the contact point was/is in use
    }
    """

    system = serializers.SerializerMethodField()
    value = serializers.SerializerMethodField()
    use = serializers.SerializerMethodField()
    rank = serializers.SerializerMethodField()
    period = serializers.SerializerMethodField()

    def get_system(self, identity):
        # Either phone | fax | email | pager | url | sms | other
        return "other"  # Todo: currently only the address is stored in the DB

    def get_value(self, identity):
        # The actual contact point details
        # Todo: since most of these fields are not required the address might be not valid
        """
        return ['{}, {}, {}, {}'.format(
            '{} ({}) {}'.format(address.street, address.aux_street,
                                address.number) if address.aux_street else '{} {}'.format(address.street,
                                                                                          address.number),
            address.postcode,
            address.city,
            address.state
        ) for address in identity.addresses]
        """
        return "not implemented"  # Todo

    def get_use(self, identity):
        return "home"

    def get_rank(self, identity):
        return (
            1  # There can only be one contact person. Therefore, this will always be 1
        )

    def get_period(self, identity):
        return ""  # Todo: This information is not beeing stored in the DB

    class Meta:
        model = Identity
        fields = [
            "system",
            "value",
            "use",
            "rank",
            "period",
        ]


class AddressSerializer(serializers.ModelSerializer):
    """
    {
      "use" : "<code>", // home | work | temp | old | billing - purpose of this address
      "type" : "<code>", // postal | physical | both
      "text" : "<string>", // Text representation of the address
      "line" : ["<string>"], // Street name, number, direction & P.O. Box etc.
      "city" : "<string>", // Name of city, town etc.
      "district" : "<string>", // District name (aka county)
      "state" : "<string>", // Sub-unit of country (abbreviations ok)
      "postalCode" : "<string>", // Postal code for area
      "country" : "<string>", // Country (e.g. can be ISO 3166 2 or 3 letter code)
      "period" : { Period } // Time period when address was/is in use
    }
    """

    use = serializers.SerializerMethodField()
    type = serializers.SerializerMethodField()
    text = serializers.SerializerMethodField()
    line = serializers.SerializerMethodField()
    city = serializers.SerializerMethodField()
    district = serializers.SerializerMethodField()
    state = serializers.SerializerMethodField()
    postalCode = serializers.SerializerMethodField()
    country = serializers.SerializerMethodField()
    period = serializers.SerializerMethodField()

    def get_use(self, address):
        return "home"  # Todo: Should all addresses have the use 'home'?

    def get_type(self, address):
        return "both"  # Todo: Should all addresses have the type 'both'?

    def get_text(self, address):
        return address  # Todo: implement __str__()-method from address

    def get_line(self, address):
        # Street name, number, direction & P.O. Box etc.
        # Todo: What is meant with directions?
        # Todo: add P.O.-box
        return [address.street, address.number]

    def get_city(self, address):
        return address.city

    def get_disctrict(self, address):
        # District name (aka county)
        return ""  # Todo: What would that be in our case?

    def get_state(self, address):
        return address.state.name

    def get_postalCode(self, address):
        return str(address.postcode)

    def get_country(self, address):
        # Country (e.g. can be ISO 3166 2 or 3 letter code)
        return address.state.code

    def get_period(self, address):
        return ""  # Todo: This information is not beeing stored in the DB

    class Meta:
        model = Address
        fields = [
            "use",
            "type",
            "text",
            "line",
            "city",
            "district",
            "state",
            "postalCode",
            "country",
            "period",
        ]


class PatientSerializer(serializers.ModelSerializer):
    resourceType = serializers.SerializerMethodField()
    identifier = serializers.SerializerMethodField()
    active = serializers.SerializerMethodField()
    name = serializers.SerializerMethodField()
    telecom = serializers.SerializerMethodField()
    gender = serializers.SerializerMethodField()
    birthdate = serializers.SerializerMethodField()
    deceasedBoolean = serializers.SerializerMethodField()
    deceasedDateTime = serializers.SerializerMethodField()
    address = serializers.SerializerMethodField()
    maritalStatus = serializers.SerializerMethodField()
    multipleBirthBoolean = serializers.SerializerMethodField()
    multipleBirthInteger = serializers.SerializerMethodField()
    photo = serializers.SerializerMethodField()
    contact = serializers.SerializerMethodField()
    communication = serializers.SerializerMethodField()
    generalPractitioner = serializers.SerializerMethodField()
    managingOrganization = serializers.SerializerMethodField()
    link = serializers.SerializerMethodField()

    def get_resourceType(self, patient):
        return "Patient"

    def get_identifier(self, patient):
        """
        The identifier should have the following structure. See http://build.fhir.org/datatypes.html#Identifier for
        details.
        {
            "use" : "<code>", // usual | official | temp | secondary | old (If known)
            "type" : { CodeableConcept }, // Description of identifier
            "system" : "<uri>", // The namespace for the identifier value
            "value" : "<string>", // The value that is unique
            "period" : { Period }, // Time period when id is/was valid for use
            "assigner" : { Reference(Organization) } // Organization that issued id (may be just text)
        }
        """
        return "not implemented"  # Todo

    def get_active(self, patient):
        # Todo: Should return a boolean stating whether this patient's record is in active use or not
        return "not implemented"

    def get_name(self, patient):
        # The patient and the name object share the same identity.
        return NameSerializer(patient.identity.name_set.all(), many=True).data

    def get_telecom(self, patient):
        """
        A contact detail for the individual. See http://build.fhir.org/datatypes.html#ContactPoint for details.
        """
        return [ContactPointSerializer(patient.identity).data]

    def get_gender(self, patient):
        # Todo: label is a CharField and can therefore be anything.
        # Todo: But gender must be one of the following: male | female | other | unknown
        return patient.identity.gender.label

    def get_birthdate(self, patient):
        return patient.identity.dob.date()

    def get_deceasedBoolean(self, patient):
        return True if patient.identity.deceased else False

    def get_deceasedDateTime(self, patient):
        return patient.identity.deceased

    def get_address(self, patient):
        return AddressSerializer(patient.identity.addresses, many=True).data

    def get_maritalStatus(self, patient):
        return (
            patient.identity.marital_status.name
            if patient.identity.marital_status
            else ""
        )

    def get_multipleBirthBoolean(self, patient):
        return "not implemented"  # Todo

    def get_multipleBirthInteger(self, patient):
        return "not implemented"  # Todo

    def get_photo(self, patient):
        return "not implemented"  # Todo

    def get_contact(self, patient):
        return "not implemented"  # Todo

    def get_communication(self, patient):
        # Todo
        return [
            {
                "language": "not implemented",
                "preferred": "not implemented",
            }
        ]

    def get_generalPractitioner(self, patient):
        return "not implemented"  # Todo

    def get_managingOrganization(self, patient):
        return "not implemented"  # Todo

    def get_link(self, patient):
        return [
            {
                "other": "not implemented",  # Todo: The other patient or related person resource that the link refers to
                "type": "not implemented",  # Todo: replaced-by | replaces | refer | seealso - type of link
            }
        ]

    class Meta:
        model = Patient
        fields = [
            "resourceType",
            "identifier",
            "active",
            "name",
            "telecom",
            "gender",
            "birthdate",
            "deceasedBoolean",
            "deceasedDateTime",
            "address",
            "maritalStatus",
            "multipleBirthBoolean",
            "multipleBirthInteger",
            "photo",
            "contact",
            "communication",
            "generalPractitioner",
            "managingOrganization",
            "link",
        ]



from enum import Enum

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..fields import CodeField, PeriodField, QuantityField, RatioField
from . import Request
from .organizations import HealthServiceProvider
from .fhir import CodeValueSet, EpisodeOfCare
from medux.common.models import BaseModel
from .datapacks import PackageDataModel


class MedicationStatus(CodeValueSet):
    pass
    # comment = models.CharField(max_length=255, blank=True)

    class Meta:
        verbose_name_plural = _("Medication Statuses")


class MedicationRoute(PackageDataModel):
    pass


# https://www.hl7.org/fhir/dosage.html
class MedicationMethod(CodeValueSet):
    pass


class Dosage(models.Model):
    class DoseAndRateType(Enum):
        CALCULATED = "calculated"
        ORDERED = "ordered"

    # sequence
    text = models.CharField(
        max_length=255, help_text=_("Free text dosage instructions e.g. '1x/day'")
    )
    additional_instructions = models.CharField(
        max_length=255,
        help_text=_(
            "Supplemental instruction or warnings to the patient - e.g. 'with meals', 'may cause drowsiness'"
        ),
        blank=True,
    )
    patient_instructions = models.CharField(
        max_length=255,
        help_text=_("Instructions in terms that are understood by the patient"),
        blank=True,
    )

    as_needed = models.BooleanField(
        default=False, help_text="Take 'as needed' (for ...)"
    )
    # as_needed_concept : e.g. "Headache"
    # could be a CharField too.

    route = models.ForeignKey(MedicationRoute, on_delete=models.PROTECT)
    # maxDosePerPeriod

    method = models.ForeignKey(
        MedicationMethod,
        blank=True,
        null=True,
        help_text=_("Technique for administering medication"),
        on_delete=models.SET_NULL,
    )

    dose_type = CodeField(
        terminology_binding=DoseAndRateType, default=DoseAndRateType.ORDERED
    )
    # dose_range = models.RangeField()
    dose_quantity = QuantityField(max_digits=10, decimal_places=2, blank=True)

    # rate ratio
    dose_rate = RatioField(
        num_max_digits=10,
        num_decimal_places=2,
        denom_max_digits=10,
        denom_decimal_places=2,
        blank=True,
    )

    def __str__(self):
        return self.text


class Ingredient(models.Model):  # PackageDataModel
    pass


class MedicationTerminologySystem(PackageDataModel):
    """A Medication terminology system like SNOMED, Austrian PZN (Pharmazentralnummer), etc."""

    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class MedicationForm(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


# https://www.hl7.org/fhir/medication.html
class Medication(PackageDataModel):
    """Definition of a medication for the purposes of prescribing, dispensing, and administering."""

    class Status(Enum):
        ACTIVE = "active"
        INACTIVE = "inactive"
        ENTERED_IN_ERROR = "entered_in_error"

    code_system = models.ForeignKey(
        MedicationTerminologySystem,
        help_text=_("The defining coding system"),
        on_delete=models.PROTECT,
    )
    code = models.CharField(
        max_length=64,
        help_text=_("The medication code within this coding system"),
    )

    status = CodeField(
        terminology_binding=Status,
        default=Status.ACTIVE,
    )

    manufacturer = models.ForeignKey(
        HealthServiceProvider,
        on_delete=models.PROTECT,
        help_text=_("Manufacturer of the medication product. (Not the distributor.)"),
    )

    # e.g. powder | tablets | capsule
    form = models.ForeignKey(MedicationForm, on_delete=models.PROTECT)

    # Amount
    # TODO: amount nominator/denominator for ratios?
    # see https://www.hl7.org/fhir/medication.html, https://www.hl7.org/fhir/datatypes.html#Ratio and
    # https://www.hl7.org/fhir/datatypes.html#Quantity
    amount_value = models.DecimalField(
        decimal_places=3, max_digits=10, help_text=_("Amount of drug in package")
    )
    amount_unit = models.CharField(max_length=25, help_text=_("Unit representation"))
    # TODO: use FK instead of code?
    amount_unit_code = models.CharField(max_length=255)

    ingredients = models.ManyToManyField(Ingredient)

    # Batch / Package information (optional)
    batch_lot_number = models.CharField(
        max_length=255, blank=True, help_text=_("Identifier assigned to batch")
    )
    batch_expiration_date = models.DateField(
        blank=True, help_text=_("When batch will expire")
    )


# MedicationStatement is not needed ATM
#
# https://www.hl7.org/fhir/medicationstatement.html
# class MedicationStatement(BaseModel):
#     """Basically a line in a list of prescribed medications.
#
#     The MedicationStatement resource is used to record medications or substances that ``the patient reports``
#     as being taken, not taking, have taken in the past or may take in the future.
#     It can also be used to record medication use that is derived from other records such as a MedicationRequest.
#
#     The statement is not used to request or order a medication. When requesting medication, when there is
#     a patient focus or instructions regarding their use, a MedicationRequest, SupplyRequest or DeviceRequest
#     should be used instead
#
#     """
#
# class Status(Enum):
#     active = _("active")
#     completed = _("completed")
#     entered_in_error = _("entered in error")
#     intended = _("intended")
#     stopped = _("stopped")
#     on_hold = _("on-hold")
#     unknown = _("unknown")
#     not_taken = _("not-taken")
#
#     status = CodeField(terminology_binding=Status)
#     patient = models.ForeignKey("Patient", on_delete=models.CASCADE)
#     medication = models.ForeignKey(Medication, on_delete=models.PROTECT)
#     encounter = models.ForeignKey(
#         "Encounter",
#         help_text=_("Encounter associated with MedicationStatement"),
#         on_delete=models.PROTECT,
#     )
#     episode = models.ForeignKey(
#         EpisodeOfCare,
#         help_text=_("EpisodeOfCare associated with MedicationStatement"),
#         on_delete=models.PROTECT,
#     )
#     note = models.TextField()


# http://hl7.org/fhir/ValueSet/medicationrequest-course-of-therapy
class CourseOfTherapy(CodeValueSet):
    # Medication request course of therapy codes:
    # http://terminology.hl7.org/CodeSystem/medicationrequest-course-of-therapy
    # "continuous|acute|seasonal"
    pass


# https://www.hl7.org/fhir/medicationrequest.html
class MedicationRequest(Request):
    """Orders for medications for a patient.

    Basically a "prescription" of one medication for outpatients."""

    class Status(Enum):
        ACTIVE = "active"
        ON_HOLD = "on_hold"
        CANCELLED = "cancelled"
        COMPLETED = "completed"
        ENTERED_IN_ERROR = "entered_in_error"
        STOPPED = "stopped"
        DRAFT = "draft"
        UNKNOWN = "unknown"

    class Priority(Enum):
        ROUTINE = "routine"  # normal priority
        URGENT = "urgent"  # higher priority than routine
        ASAP = "asap"  # as soon as possible
        IMMEDIATELY = "stat"  # immediately, emergency

    class Intents(Enum):
        PROPOSAL = "proposal"
        PLAN = "plan"
        ORDER = "order"
        ORIGINAL_ORDER = "original-order"
        REFLEX_ORDER = "reflex-order"
        FILLER_ORDER = "filler-order"
        INSTANCE_ORDER = "instance-order"
        OPTION = "option"

    status = CodeField(terminology_binding=Status)
    # status_reason = models.ForeignKey(MedicationRequestStatusReason, on_delete=models.PROTECT)

    # Whether the request is a proposal, plan, or an original order
    intent = CodeField(terminology_binding=Intents)

    # inpatient, outpatient, community, discharge
    # see https://www.hl7.org/fhir/codesystem-medicationrequest-category.html
    # category = models.ForeignKey(MedicationRequestCategory, on_delete=models.PROTECT)

    priority = CodeField(terminology_binding=Priority)
    do_not_perform = models.BooleanField(
        default=False, help_text=_("True if request is prohibiting action")
    )

    reported = models.BooleanField(default=False)

    medication = models.ForeignKey(Medication, on_delete=models.CASCADE)

    subject = models.ForeignKey(
        "Patient",
        on_delete=models.CASCADE,
        related_name="+",
        help_text=_("Who medication request is for"),
    )

    encounter = models.ForeignKey("Encounter", on_delete=models.CASCADE)

    # = created
    # authored_on = models.DateTimeField(auto_now=True)

    # TODO: requester: Person/Organization/Device
    requester = models.ForeignKey(
        HealthServiceProvider,
        on_delete=models.SET_NULL,
        null=True,
        help_text=_("Who requested the request"),
    )

    # default: current user
    recorder = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    # reason_code, reason_reference ...

    based_on = models.ForeignKey("CarePlan", null=True, on_delete=models.SET_NULL)

    # priorPrescription
    replaces = models.ForeignKey(
        "MedicationRequest",
        on_delete=models.PROTECT,
        help_text=_("An order/prescription that is being replaced"),
        related_name="replaced_by",
    )

    course_of_therapy_type = models.ForeignKey(
        CourseOfTherapy, on_delete=models.PROTECT, blank=True
    )

    note = models.CharField(
        max_length=255, help_text=_("Information about the prescription")
    )

    dosage_instruction = models.OneToOneField(Dosage, on_delete=models.PROTECT)


class CarePlan(BaseModel):
    """based upon the FHIR "CarePlan": https://hl7.org/fhir/careplan.html"""

    class Status(Enum):
        DRAFT = "draft"
        ACTIVE = "active"
        ON_HOLD = "on-hold"
        REVOKED = "revoked"
        COMPLETED = "completed"
        ENTERED_IN_ERROR = "entered_in_error"
        UNKNOWN = "unknown"

    class Intents(Enum):
        PROPOSAL = "proposal"
        PLAN = "plan"
        ORDER = "order"
        OPTION = "option"

    # based_on = models.ForeignKey(
    #     "CarePlan",
    #     null=True,
    #     on_delete=models.SET_NULL,
    #     help_text=_("Fulfills CarePlan"),
    #     related_name="parent",
    # )

    # part_of = models.ForeignKey(
    #     "CarePlan",
    #     null=True,
    #     on_delete=models.SET_NULL,
    #     help_text=_("Part of referenced CarePlan"),
    # )

    status = CodeField(terminology_binding=Status)
    intent = CodeField(terminology_binding=Intents)

    replaces = models.ForeignKey(
        "CarePlan",
        null=True,
        on_delete=models.SET_NULL,
        help_text=_("CarePlan replaced by this CarePlan"),
        related_name="replaced_by",
    )

    # TODO: could be also Patient, RelatedPerson etc...
    author = models.ForeignKey(
        HealthServiceProvider, null=True, on_delete=models.SET_NULL
    )
    title = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Human-friendly name for the care plan"),
    )
    description = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Summary of plan"),
    )
    subject = models.ForeignKey("Patient", on_delete=models.CASCADE)

    # TODO: keep null=True? CarePlan is possible "dangling" without encounter?
    encounter = models.ForeignKey("Encounter", null=True, on_delete=models.SET_NULL)

    period = PeriodField()

    # careTeam, addresses, supportingInfo, goal, activity,

    note = models.TextField(blank=True, help_text=_("Comments about the plan"))


class MedicationPlan(CarePlan):
    """A medication plan that can be printed etc.

    Contains a list of medication requests, and could be printed, handed out to the patient, or given to another
    health care service worker.
    """

    medications = models.ManyToManyField(MedicationRequest)


# class DeviceRequest(Request):
#     pass


# https://www.hl7.org/fhir/medicationadministration.html
# class MedicationAdministration(BaseModel):
#     """Describes the event of a patient consuming or otherwise being administered a medication.
#     This may be as simple as swallowing a tablet or it may be a long running infusion.
#
#     Related resources tie this event to the authorizing prescription, and the specific encounter
#     between patient and health care practitioner."""



# no global imports of models,
# so this functions can be used before apps are loaded.
from typing import Any

from django.http import HttpRequest

PreferencesContext = dict[str, Any]


def is_editable(setting, request: HttpRequest):
    from medux.preferences.definitions import Scope

    match setting.scope:
        case Scope.VENDOR:
            return False
        case Scope.TENANT:
            return (
                request.user.has_perm("preferences.change_own_tenant_scopedpreference")
                and setting.tenant == request.user.tenant
            )
        case Scope.USER:
            return request.user.has_perm("preferences.change_own_user_scopedpreference")
        case Scope.GROUP:
            return (
                request.user.has_perm("preferences.change_own_group_scopedpreference")
                and setting.group in request.user.groups.all()
            )
        case Scope.DEVICE:
            return request.user.has_perm("preferences.change_device_scopedpreference")


def is_deletable(setting, request: HttpRequest):
    return is_editable(setting, request)


def is_effective(setting, request: HttpRequest):
    return setting == get_effective_setting(setting.namespace, setting.key, request)


def get_effective_setting(namespace: str, key: str, request):
    """
    :return: the effective ScopedPreference with the give namespace/key.
    """
    from .models import ScopedPreference
    from medux.preferences.definitions import Scope

    queryset = ScopedPreference.objects.filter(
        base__namespace=namespace, base__key=key
    ).order_by("-scope")

    # traverse preferences in reverse Order (USER -> VENDOR),
    # and take the first matching
    for item in queryset:
        scope = item.scope

        if scope == Scope.USER and request.user == item.user:
            return item

        if scope == Scope.DEVICE and request.device == item.device:
            return item

        if scope == Scope.GROUP and item.group in request.user.groups.all():
            return item

        if scope == Scope.TENANT:
            # for authenticated users, the user's tenant counts!
            if request.user.is_authenticated:
                if request.user.tenant == item.tenant:
                    return item
            # for anonymous users, the site tenant is the valid one.
            else:
                if request.site.tenant == item.tenant:
                    return item

        if scope == Scope.VENDOR:
            return item



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from medux.common.api.interfaces import IMenuItem
from medux.core.api import IViewMode


class SearchViewMode(IViewMode):
    title = _("Search")
    url = reverse("patient_list")
    icon = "bi-search"
    weight = 20


# Patient menu
class PatientNew(IMenuItem):
    menu = "main_menu"
    title = _("New")
    url = reverse("patient_new")
    weight = 0
    icon = "person-circle"


class Patient(IMenuItem):
    title = _("Patient")
    url = "/"
    weight = 0
    children = [
        PatientNew,
        # MenuSeparator,
    ]


class ExtrasPreferences(IMenuItem):
    _("Preferences")
    reverse("preferences")
    slug = "preferences"
    weight = 10
    icon = "gear"
    icon_only = True  # TODO: implement icon_only


class Extras(IMenuItem):
    menu = "main_menu"
    _("Extras")
    url = reverse("home")
    weight = 30
    children = [ExtrasPreferences]


# top right menu
class Notifications(IMenuItem):
    menu = ("top_navbar",)
    _("Notifications"),
    url = (reverse("home"),)
    weight = (20,)
    icon = ("bi-bell",)
    icon_only = (True,)
    badge = (True,)  # FIXME: this shouldn't be hardcoded here


# class Home(IMenuItem):
#     menu="top_navbar"
#
#     MenuItem(
#         title=lambda request: request.user,
#         url=reverse("home"),
#         slug="myaccount",
#         weight=99,
#         icon="bi-user",
#         children=[
#             MenuItem("Edit Profile", url=reverse("home"), icon="bi-user"),
#             MenuItem(
#                 title="Admin",
#                 url=reverse("admin:index"),
#                 # check=lambda request: request.user.is_superuser,
#             ),
#             MenuSeparator(),
#             MenuItem(
#                 title=_("Logout"),
#                 url=reverse("logout"),
#                 icon="bi-box-arrow-right",
#             ),
#         ],
#     ),
# )



from django import apps
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from medux.core.models import Patient
from medux.common.models import BaseModel


class TaskList(models.Model):
    """This is a "continuous task list"""

    name = models.CharField(max_length=255)

    class Meta:
        verbose_name = _("Task List")
        verbose_name_plural = _("Task Lists")

    def __str__(self) -> str:
        return str(self.name)

    @property
    def is_complete(self) -> bool:
        return not self.tasks.filter(is_done=False).exists()

    @property
    def complete_tasks(self) -> models.QuerySet["CommonTask"]:
        return self.tasks.filter(is_done=True)

    @property
    def incomplete_tasks(self) -> models.QuerySet["CommonTask"]:
        return self.tasks.filter(is_done=False)


class CommonTask(BaseModel):
    task_list = models.ForeignKey(
        TaskList, on_delete=models.PROTECT, related_name="tasks"
    )
    name = models.CharField(max_length=255)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    is_done = models.BooleanField(default=False)

    class Meta:
        verbose_name = _("Task Item")
        verbose_name_plural = _("Task Items")

    def __str__(self) -> str:
        return str(self.name)


class PatientTask(CommonTask):
    """Tasks related to a patient"""

    patient = models.ForeignKey(Patient, on_delete=models.PROTECT, parent_link=True)


# ---------------------------------------------------------------------------
# Ideas:

# class DeviceTask(CommonTask):
#     """Tasks related to a"""
#
#     device = models.ForeignKey(Device, on_delete=models.PROTECT, parent_link=True)


# class DiagnosticReportTask(CommonTask):
#     """Tasks related to a diagnostic report.
#
#     This task is generated automatically then creating a referral to another Health care service.
#     It should be possible to keep track of it using this task.
#     Incoming diagnostic results create this Task too, so you can easily keep an eye on them."""
#
#     report = models.ForeignKey(..., on_delete=models.PROTECT, parent_link=True)



import logging

from django import template
from django.urls import reverse
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

logger = logging.getLogger(__file__)
register = template.Library()


def render_crumb(title, url=None, link=True):
    """Helper function"""
    title = escape(title)
    if url:
        crumb = f'<li class="breadcrumb-item"><a href="{url}">{title}</a></li>'
    else:
        crumb = f'<li class="breadcrumb-item active" aria-current="page">{title}</li>'

    return crumb


@register.simple_tag(takes_context=True)
def breadcrumb(context, title: str, view_name: str = "", path: str = "", **kwargs):
    """Renders a breadcrumb.

    Usage:
        {% breadcrumb "Start" "home" %}
        {% breadcrumb context_var  url_var %}


    :param title str: The title of the breadcrumb. This string is
        translated within the templatetag.
    :param view_name str: The (optional) URL name of the breadcrumb, which is resolved automatically.
    :param path str: The raw path which is used directly as link

    TODO: use the context to match current page -> don't show as link
    """
    # if title is an object, use its __str__:
    title = str(title)

    link = True
    if view_name:
        url = reverse(view_name, kwargs=kwargs)
        # link = url == context["sd"]
    elif path:
        url = path
    else:
        raise AttributeError(
            "Either 'url' or 'path' attribute must be provided in 'breadcrumb' templatetag."
        )

    return mark_safe(render_crumb(_(title), url, link))



# Generated by Django 4.1.3 on 2022-12-07 22:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0002_create_fixtures"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="avatar",
            field=models.ImageField(null=True, upload_to="user_avatars/"),
        ),
    ]



from gdaps.api import Interface


@Interface
class IViewMode:
    """Interface for MedUX ViewModes.

    A ViewMode provides data and methods for displaying an icon for the
    main viewing modes of MedUX.
    """

    __service__ = True

    title = ""
    """The title which is displayed when hovering over the icon."""

    url = ""
    """the URL that should be called. Can be a Dajngo reverse() URL."""

    icon = "bi-file-earmark"
    """the icon as text, using Bootstrap Icons"""

    icon_type = "bi"  # png
    """icon_type determines where the icon is fetched from:
        bi (Bootstrap icon) or png, then it is loaded from static files"""

    weight = 0
    """the weight of the icon in the list. The higher the weight, the "deeper" the icon."""


@Interface
class IGlobalJavascript:
    """Interface for adding Js file to a global scope.

    The given file will be loaded in the global context and is then available to
    all other plugins too, in every loaded page.
    Be aware just to add Js code that is small and fast, to not blow up the application
    as whole.

    You have to specify the (relative to static dir) file path of the Js script in the
    `file` attribute.
    """

    __service__ = True

    file: str = ""
    """The (relative to static dir) file path of the Js script"""


@Interface
class ICommand:
    """A CommandLine command that executes a defined function on a shortcut.

    A written command could be ``m dicl 75`` what means the "m" command with could be parsed
    as "medication" command with a search for "Diclofenac 75mg Tbl"."""

    shortcut: str = None
    description: str = None

    def execute(self, *args, **kwargs):
        raise NotImplementedError
